package gov.usgs.earthquake.nshmp.calc;

import static com.google.common.base.Preconditions.checkNotNull;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkInRange;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P01;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P5;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA10P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA2P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA3P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA4P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA5P0;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA7P5;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Writer;
import java.lang.reflect.Type;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Set;
import java.util.TreeMap;

import com.google.common.collect.Maps;
import com.google.common.collect.Range;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonDeserializationContext;
import com.google.gson.JsonDeserializer;
import com.google.gson.JsonElement;
import com.google.gson.JsonParseException;
import com.google.gson.JsonPrimitive;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;

import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.gmm.UsgsDampingScaling;
import gov.usgs.earthquake.nshmp.gmm.UsgsVerticalMotions;
import gov.usgs.earthquake.nshmp.model.SourceType;
import gov.usgs.earthquake.nshmp.model.TectonicSetting;

/**
 * Calculation configuration. Static factory methods
 * {@link #copyOf(CalcConfig)}, {@link #from(Path)}, and {@link #defaults()}
 * each return a builder for possible additional config modification.
 *
 * @author U.S. Geological Survey
 */
public final class CalcConfig {

  private static final String DEFAULTS = "/calc/calc-config-defaults.json";

  /**
   * The expected calculation configuration file name when included in source
   * models, {@code calc-config.json}.
   */
  public static final String FILE_NAME = "calc-config.json";

  /** Hazard calculation configuration. */
  public final Hazard hazard;

  /** Ground motion model configuration. */
  public final Gmm gmm;

  /** Disaggregation configuration. */
  public final Disagg disagg;

  /** Earthquake rate configuration. */
  public final Rate rate;

  /** Output configuration. */
  public final Output output;

  /** Performance and optimization configuration. */
  public final Performance performance;

  private CalcConfig(Builder builder) {
    hazard = builder.hazard.build();
    gmm = builder.gmm.build();
    disagg = builder.disagg.build();
    rate = builder.rate.build();
    output = builder.output.build();
    performance = builder.performance.build();
  }

  /**
   * Initialize a new builder with values copied from the supplied config.
   */
  public static Builder copyOf(CalcConfig config) {
    Builder b = new Builder();
    b.hazard.copy(config.hazard);
    b.gmm.copy(config.gmm);
    b.disagg.copy(config.disagg);
    b.rate.copy(config.rate);
    b.output.copy(config.output);
    b.performance.copy(config.performance);
    return b;
  }

  /**
   * Create a new builder from the resource at the specified {@code path}. This
   * factory constructor only sets those fields that are explicitely defined in
   * the supplied resource.
   *
   * @param path to configuration file
   * @throws IOException
   */
  public static Builder from(Path path) throws IOException {
    try (Reader reader = Files.newBufferedReader(path)) {
      return GSON.fromJson(reader, Builder.class);
    }
  }

  /**
   * Create a new builder with all fields initialized to default values.
   */
  public static Builder defaults() throws IOException {
    InputStream is = CalcConfig.class.getResourceAsStream(DEFAULTS);
    try (Reader reader = new InputStreamReader(is)) {
      return GSON.fromJson(reader, Builder.class);
    }
  }

  /*
   * Developer notes:
   *
   * Need to add check(s) that array valued properties are not empty, which will
   * have different behavior than null.
   */

  /**
   * Hazard calculation configuration.
   */
  public final static class Hazard {

    /**
     * The probability distribution model to use when computing hazard curves.
     *
     * <p><b>Default:</b> {@link ExceedanceModel#TRUNCATION_UPPER_ONLY}
     */
    public final ExceedanceModel exceedanceModel;

    /**
     * The number of standard deviations (σ) at which to truncate a
     * distribution. This field is ignored if an {@link ExceedanceModel} does
     * not implement truncation.
     *
     * <p><b>Default:</b> {@code 3.0}
     */
    public final double truncationLevel;

    /**
     * The set of IMTs for which calculations should be performed.
     *
     * <p><b>Default:</b> Multi-period response spectral accelerations plus PGV
     *
     * <pre>
     * [
     *   PGA, PGV,
     *   0.01s, 0.02s, 0.03s, 0.05s, 0.075s,
     *   0.1s, 0.15s, 0.2s, 0.25s, 0.3s, 0.4s, 0.5s, 0.75s,
     *   1.0s, 1.5s, 2.0s, 3.0s, 4.0s, 5.0s, 7.5s,
     *   10.0s
     * ]}
     * </pre>
     */
    public final Set<Imt> imts;

    /**
     * The set of tectonic settings for which calculations should be performed.
     * If empty or {@code null} in a config file, then all tectonic settings are
     * considered, otherwise the source model will be filtered to only consider
     * those listed.
     *
     * <p><b>Default:</b> []
     */
    public final Set<TectonicSetting> tectonicSettings;

    /**
     * The set of source types for which calculations should be performed. If
     * empty or {@code null} in a config file, then all source types are
     * considered, otherwise the source model will be filtered to only consider
     * those listed.
     *
     * <p><b>Default:</b> []
     */
    public final Set<SourceType> sourceTypes;

    /**
     * The set of Vs30s (average shear-wave velocity down to 30 meters depth)
     * for which calculations should be performed. If empty or {@code null} in a
     * config file, then Vs30 will be derived from a users site file or the
     * {@link SiteDefaults#vs30} value.
     *
     * <p><b>Default:</b> []
     */
    public final Set<Double> vs30s;

    /**
     * Whether or not consider site-specific data that may accompany a model.
     * Set to false to turn off deep basin scaling effects, for example.
     *
     * <p><b>Default:</b> true
     */
    public final boolean useSiteData;

    /**
     * The value format for hazard curves.
     *
     * <p><b>Default:</b> {@link ValueFormat#ANNUAL_RATE}
     */
    public final ValueFormat valueFormat;

    private final Map<Imt, double[]> customImls;

    /* Do not serialize to JSON */
    private final transient Map<Imt, XySequence> modelCurves;
    private final transient Map<Imt, XySequence> logModelCurves;

    private Hazard(Builder b) {
      this.exceedanceModel = b.exceedanceModel;
      this.truncationLevel = b.truncationLevel;
      this.imts = b.imts;
      this.tectonicSettings = b.tectonicSettings;
      this.sourceTypes = b.sourceTypes;
      this.vs30s = b.vs30s;
      this.useSiteData = b.useSiteData;
      this.valueFormat = b.valueFormat;

      this.customImls = b.customImls;
      this.modelCurves = b.curveMap;
      this.logModelCurves = b.logCurveMap;
    }

    /**
     * An immutable map of model curves where x-values are in linear space. In
     * the absence of specific values being provided in a config file, the
     * following defaults are used:
     *
     * <p><b>{@code PGA}</b> and <b>{@code SA}</b>: {@code [0.0025,
     * 0.0045, 0.0075, 0.0113, 0.0169, 0.0253, 0.0380, 0.0570, 0.0854, 0.128,
     * 0.192, 0.288, 0.432, 0.649, 0.973, 1.46, 2.19, 3.28, 4.92, 7.38]}
     *
     * <p><b>{@code PGV}</b>: {@code [0.0100, 0.0177, 0.0312, 0.0552,
     * 0.0976, 0.173, 0.305, 0.539, 0.953, 1.68, 2.98, 5.26, 9.30, 16.4,
     * 29.1, 51.3, 90.8, 160, 284, 501]}
     */
    Map<Imt, XySequence> modelCurves() {
      return modelCurves;
    }

    /**
     * An immutable map of model curves where x-values are in natural-log space.
     * @see #modelCurves() linear space defaults
     */
    Map<Imt, XySequence> logModelCurves() {
      return logModelCurves;
    }

    private static final class Builder {

      ExceedanceModel exceedanceModel;
      Double truncationLevel;
      Set<Imt> imts;
      Set<TectonicSetting> tectonicSettings;
      Set<SourceType> sourceTypes;
      Set<Double> vs30s;
      Boolean useSiteData;
      ValueFormat valueFormat;
      Map<Imt, double[]> customImls;
      Map<Imt, XySequence> curveMap;
      Map<Imt, XySequence> logCurveMap;

      Hazard build() {
        initCurveMaps();
        checkNotNull(exceedanceModel);
        checkNotNull(truncationLevel);
        checkNotNull(imts);
        checkNotNull(tectonicSettings);
        checkNotNull(sourceTypes);
        checkNotNull(vs30s);
        checkNotNull(useSiteData);
        checkNotNull(valueFormat);
        checkNotNull(customImls);
        return new Hazard(this);
      }

      void copy(Hazard that) {
        this.exceedanceModel = that.exceedanceModel;
        this.truncationLevel = that.truncationLevel;
        this.imts = that.imts;
        this.tectonicSettings = that.tectonicSettings;
        this.sourceTypes = that.sourceTypes;
        this.vs30s = that.vs30s;
        this.useSiteData = that.useSiteData;
        this.valueFormat = that.valueFormat;
        this.customImls = that.customImls;
      }

      void extend(Builder that) {
        if (that.exceedanceModel != null) {
          this.exceedanceModel = that.exceedanceModel;
        }
        if (that.truncationLevel != null) {
          this.truncationLevel = that.truncationLevel;
        }
        if (that.imts != null) {
          this.imts = that.imts;
        }
        if (that.tectonicSettings != null) {
          this.tectonicSettings = that.tectonicSettings;
        }
        if (that.sourceTypes != null) {
          this.sourceTypes = that.sourceTypes;
        }
        if (that.vs30s != null) {
          this.vs30s = that.vs30s;
        }
        if (that.useSiteData != null) {
          this.useSiteData = that.useSiteData;
        }
        if (that.valueFormat != null) {
          this.valueFormat = that.valueFormat;
        }
        if (that.customImls != null) {
          this.customImls.putAll(that.customImls);
        }
      }

      private void initCurveMaps() {
        Map<Imt, XySequence> curveMap = Maps.newEnumMap(Imt.class);
        Map<Imt, XySequence> logCurveMap = Maps.newEnumMap(Imt.class);
        for (Imt imt : imts) {
          double[] imls = customImls.containsKey(imt)
              ? customImls.get(imt)
              : imlsFor(imt);
          // use a copy so conversion to log doesn't modify internal
          imls = Arrays.copyOf(imls, imls.length);
          curveMap.put(imt, XySequence.create(imls, new double[imls.length]));
          DoubleData.ln(imls);
          logCurveMap.put(imt, XySequence.create(imls, new double[imls.length]));
        }
        this.curveMap = curveMap;
        this.logCurveMap = logCurveMap;
      }
    }

    private static double[] imlsFor(Imt imt) {
      if (imt.isSA() || imt == Imt.PGA) {
        return imlMap.ceilingEntry(imt).getValue();
      } else if (imt == Imt.PGV) {
        return IMLS_PGV;
      } else {
        /* IMLs not yet assigned for IMT. */
        throw new UnsupportedOperationException();
      }
    }

    /*
     * IML defaults. These were assigned in nshm-conus-2018 and set as custom
     * overrides. We now include them as built-in default values. For any
     * periods in between those explicitely specified, values for the next
     * highest supported period is used (see navigable imlMap, below). Users
     * need only specify overriding IMLs in the "imls" hazaard config member.
     */
    static final double[] IMLS_SA0P01 = {
        0.00233, 0.00350, 0.00524, 0.00786, 0.0118, 0.0177, 0.0265, 0.0398, 0.0597, 0.0896, 0.134,
        0.202, 0.302, 0.454, 0.680, 1.02, 1.53, 2.30, 3.44, 5.17 };
    static final double[] IMLS_SA0P02 = {
        0.00283, 0.00424, 0.00637, 0.00955, 0.0143, 0.0215, 0.0322, 0.0483, 0.0725, 0.109, 0.163,
        0.245, 0.367, 0.551, 0.826, 1.24, 1.86, 2.79, 4.18, 6.27 };
    static final double[] IMLS_SA0P5 = {
        0.00333, 0.00499, 0.00749, 0.0112, 0.0169, 0.0253, 0.0379, 0.0569, 0.0853, 0.128, 0.192,
        0.288, 0.432, 0.648, 0.972, 1.46, 2.19, 3.28, 4.92, 7.38 };
    static final double[] IMLS_SA2P0 = {
        0.00250, 0.00375, 0.00562, 0.00843, 0.0126, 0.0190, 0.0284, 0.0427, 0.0640, 0.0960, 0.144,
        0.216, 0.324, 0.486, 0.729, 1.09, 1.64, 2.46, 3.69, 5.54 };
    static final double[] IMLS_SA3P0 = {
        0.00200, 0.00300, 0.00449, 0.00674, 0.0101, 0.0152, 0.0228, 0.0341, 0.0512, 0.0768, 0.115,
        0.173, 0.259, 0.389, 0.583, 0.875, 1.31, 1.97, 2.95, 4.43 };
    static final double[] IMLS_SA4P0 = {
        0.00133, 0.00200, 0.00300, 0.00449, 0.00674, 0.0101, 0.0152, 0.0228, 0.0341, 0.0512, 0.0768,
        0.115, 0.173, 0.259, 0.389, 0.583, 0.875, 1.31, 1.97, 2.95 };
    static final double[] IMLS_SA5P0 = {
        0.000999, 0.00150, 0.00225, 0.00337, 0.00506, 0.00758, 0.0114, 0.0171, 0.0256, 0.0384,
        0.0576, 0.0864, 0.130, 0.194, 0.292, 0.437, 0.656, 0.984, 1.48, 2.21 };
    static final double[] IMLS_SA7P5 = {
        0.000499, 0.000749, 0.00112, 0.00169, 0.00253, 0.00379, 0.00569, 0.00853, 0.0128, 0.0192,
        0.0288, 0.0432, 0.0648, 0.0972, 0.146, 0.219, 0.328, 0.492, 0.738, 1.11 };
    static final double[] IMLS_SA10P0 = {
        0.000333, 0.000499, 0.000749, 0.00112, 0.00169, 0.00253, 0.00379, 0.00569, 0.00853, 0.0128,
        0.0192, 0.0288, 0.0432, 0.0648, 0.0972, 0.146, 0.219, 0.328, 0.492, 0.738 };

    static final double[] IMLS_PGV = {
        0.237, 0.355, 0.532, 0.798, 1.19, 1.80, 2.69, 4.04, 6.06, 9.09, 13.6, 20.5, 30.7,
        46.0, 69.0, 103.0, 155.0, 233.0, 349.0, 525.0 };

    /* NavigableMap facilitates assigning IMLs to ranges of spectral periods. */
    static final NavigableMap<Imt, double[]> imlMap = new TreeMap<>(Map.of(
        SA0P01, IMLS_SA0P01,
        SA0P02, IMLS_SA0P02,
        SA0P5, IMLS_SA0P5,
        SA2P0, IMLS_SA2P0,
        SA3P0, IMLS_SA3P0,
        SA4P0, IMLS_SA4P0,
        SA5P0, IMLS_SA5P0,
        SA7P5, IMLS_SA7P5,
        SA10P0, IMLS_SA10P0));
  }

  /**
   * Ground motion model configuration.
   */
  public final static class Gmm {

    /**
     * Viscous damping ratio. The default value is consistent with that used in
     * {@link GroundMotionModel} most (GMM) development. Values other than 5%
     * will apply a damping scaling factor (DSF) to adjust the 5% damped
     * spectral ordinates predicted by GMMs.
     *
     * <p><b>Default:</b> 5% <br><b>Range:</b> [0.5..30]%
     *
     * @see UsgsDampingScaling
     */
    public final double dampingRatio;

    /**
     * Factor by which the aleatory variability (sigma) for all GMMs should be
     * scaled.
     *
     * <p><b>Default:</b> 1.0<br><b>Range:</b> {@code 0.5} to {@code 1.0} (50%
     * to 100%, no scaling)
     */
    public final double sigmaScale;

    /**
     * Whether to compute vertical ground motions instead of horizontal.
     *
     * <p><b>Default:</b> false
     *
     * @see UsgsVerticalMotions
     */
    public final boolean vertical;

    private Gmm(Builder b) {
      this.dampingRatio = b.dampingRatio;
      this.sigmaScale = b.sigmaScale;
      this.vertical = b.vertical;
    }

    private static final class Builder {

      Double dampingRatio;
      Double sigmaScale;
      Boolean vertical;

      Gmm build() {
        checkNotNull(dampingRatio);
        checkInRange(
            UsgsDampingScaling.DAMPING_RATIO_RANGE,
            "dampingRatio",
            dampingRatio);
        checkNotNull(sigmaScale);
        checkInRange(
            Range.closed(0.5, 1.0),
            "sigmaScale",
            sigmaScale);
        checkNotNull(vertical);
        return new Gmm(this);
      }

      void copy(Gmm that) {
        this.dampingRatio = that.dampingRatio;
        this.sigmaScale = that.sigmaScale;
        this.vertical = that.vertical;
      }

      void extend(Builder that) {
        if (that.dampingRatio != null) {
          this.dampingRatio = that.dampingRatio;
        }
        if (that.sigmaScale != null) {
          this.sigmaScale = that.sigmaScale;
        }
        if (that.vertical != null) {
          this.vertical = that.vertical;
        }
      }
    }
  }

  /**
   * Disaggregation configuration.
   */
  public static final class Disagg {

    /**
     * The return period (in years) at which disaggregations will be performed
     * unless period-specific ground motion levels are included in a sites file.
     */
    public final Integer returnPeriod;

    /**
     * The distance, magnitude, and epsilon bins into which contributing sources
     * to hazard are sorted.
     */
    public final Bins bins;

    /**
     * The minimum contribution (in %) that a source must make to hazard to be
     * included on the contributor source list in a disaggregation result.
     */
    public final Double contributorLimit;

    private Disagg(Builder b) {
      this.returnPeriod = b.returnPeriod;
      this.bins = b.bins;
      this.contributorLimit = b.contributorLimit;
    }

    /**
     * The distance, magnitude, and epsilon bins into which contributing sources
     * to hazard will be sorted.
     */
    public static final class Bins {

      /** Minimum distance. Lower edge of smallest distance bin. */
      public final Double rMin;

      /** Maximum distance. Upper edge of largest distance bin. */
      public final Double rMax;

      /** Distance bin width. */
      public final Double Δr;

      /** Minimum magnitude. Lower edge of smallest magnitude bin. */
      public final Double mMin;

      /** Maximum magnitude. Upper edge of largest magnitude bin. */
      public final Double mMax;

      /** Magnitude bin width. */
      public final Double Δm;

      /** Minimum epsilon. Lower edge of smallest epsilon bin. */
      public final Double εMin;

      /** Maximum epsilon. Upper edge of largest epsilon bin. */
      public final Double εMax;

      /** Epsilon bin width. */
      public final Double Δε;

      // supports final fields and GSON deserilaization
      private Bins() {
        rMin = 0.0;
        rMax = 0.0;
        Δr = 0.0;
        mMin = 0.0;
        mMax = 0.0;
        Δm = 0.0;
        εMin = 0.0;
        εMax = 0.0;
        Δε = 0.0;
      }
    }

    private static final class Builder {

      Integer returnPeriod;
      Bins bins;
      Double contributorLimit;

      Disagg build() {
        checkNotNull(bins);
        checkNotNull(bins.rMin);
        checkNotNull(bins.rMax);
        checkNotNull(bins.Δr);
        checkNotNull(bins.mMin);
        checkNotNull(bins.mMax);
        checkNotNull(bins.Δm);
        checkNotNull(bins.εMin);
        checkNotNull(bins.εMax);
        checkNotNull(bins.Δε);
        checkNotNull(contributorLimit);
        return new Disagg(this);
      }

      void copy(Disagg that) {
        this.returnPeriod = that.returnPeriod;
        this.bins = that.bins;
        this.contributorLimit = that.contributorLimit;
      }

      void extend(Builder that) {
        if (that.returnPeriod != null) {
          this.returnPeriod = that.returnPeriod;
        }
        if (that.bins != null) {
          this.bins = that.bins;
        }
        if (that.contributorLimit != null) {
          this.contributorLimit = that.contributorLimit;
        }
      }
    }
  }

  /**
   * Magnitude-frequency distribution configuration.
   */
  public static final class Rate {

    /**
     * The magnitude discretization.
     */
    public final Bins bins;

    /**
     * The distance from a site within which all sources should be included.
     *
     * <p><b>Default:</b> {@code 20 km}
     */
    public final Double distance;

    /**
     * The rate data distribution type.
     *
     * <p><b>Default:</b> {@link DistributionFormat#INCREMENTAL}
     */
    public final DistributionFormat distributionFormat;

    /**
     * The value format for rate data.
     *
     * <p><b>Default:</b> {@link ValueFormat#ANNUAL_RATE}
     */
    public final ValueFormat valueFormat;

    /**
     * The timespan of interest when computing Poisson probabilities.
     *
     * <p><b>Default:</b> {@code 30 years}
     */
    public final Double timespan;

    private Rate(Builder b) {
      this.bins = b.bins;
      this.distance = b.distance;
      this.distributionFormat = b.distributionFormat;
      this.valueFormat = b.valueFormat;
      this.timespan = b.timespan;
    }

    private static final class Builder {

      Bins bins;
      Double distance;
      DistributionFormat distributionFormat;
      ValueFormat valueFormat;
      Double timespan;

      Rate build() {
        checkNotNull(bins);
        checkNotNull(bins.mMin);
        checkNotNull(bins.mMax);
        checkNotNull(bins.Δm);
        checkNotNull(distance);
        checkNotNull(distributionFormat);
        checkNotNull(valueFormat);
        checkNotNull(timespan);
        return new Rate(this);
      }

      void copy(Rate that) {
        this.bins = that.bins;
        this.distance = that.distance;
        this.distributionFormat = that.distributionFormat;
        this.valueFormat = that.valueFormat;
        this.timespan = that.timespan;
      }

      void extend(Builder that) {
        if (that.bins != null) {
          this.bins = that.bins;
        }
        if (that.distance != null) {
          this.distance = that.distance;
        }
        if (that.distributionFormat != null) {
          this.distributionFormat = that.distributionFormat;
        }
        if (that.valueFormat != null) {
          this.valueFormat = that.valueFormat;
        }
        if (that.timespan != null) {
          this.timespan = that.timespan;
        }
      }
    }

    /**
     * The mfd magnitude discretization.
     */
    public static final class Bins {

      /** Minimum magnitude. Lower edge of smallest magnitude bin. */
      public final Double mMin;

      /** Maximum magnitude. Upper edge of largest magnitude bin. */
      public final Double mMax;

      /** Magnitude bin width. */
      public final Double Δm;

      // supports final fields and GSON deserilaization
      Bins() {
        mMin = 0.0;
        mMax = 0.0;
        Δm = 0.0;
      }
    }
  }

  /**
   * Data and file output settings.
   */
  public static final class Output {

    /**
     * The directory to write any results to.
     *
     * <p><b>Default:</b> {@code "hazout"}.
     */
    public final Path directory;

    /**
     * The different {@linkplain DataType types} of data to save. Note that
     * {@link DataType#TOTAL} will <i>always</i> be included in this set,
     * regardless of any user settings.
     *
     * <p><b>Default:</b> [{@link DataType#TOTAL}]
     */
    public final Set<DataType> dataTypes;

    /**
     * The return periods to create hazard map datasets from hazard curves when
     * {@link DataType#MAP} is used.
     *
     * <p><b>Default:</b> [475, 975, 2475]
     */
    public final List<Integer> returnPeriods;

    private Output(Builder b) {
      this.directory = b.directory;
      Set<DataType> dataTypes = EnumSet.copyOf(b.dataTypes);
      dataTypes.add(DataType.TOTAL);
      dataTypes.add(DataType.MAP);
      this.dataTypes = Collections.unmodifiableSet(dataTypes);
      this.returnPeriods = List.copyOf(b.returnPeriods);
    }

    private static final class Builder {

      Path directory;
      Set<DataType> dataTypes;
      List<Integer> returnPeriods;

      Output build() {
        checkNotNull(directory);
        checkNotNull(dataTypes);
        checkNotNull(returnPeriods);
        return new Output(this);
      }

      void copy(Output that) {
        this.directory = that.directory;
        this.dataTypes = EnumSet.copyOf(that.dataTypes);
        this.returnPeriods = that.returnPeriods;
      }

      void extend(Builder that) {
        if (that.directory != null) {
          this.directory = that.directory;
        }
        if (that.dataTypes != null) {
          this.dataTypes = EnumSet.copyOf(that.dataTypes);
        }
        if (that.returnPeriods != null) {
          this.returnPeriods = that.returnPeriods;
        }
      }
    }
  }

  /**
   * Performance and optimization settings.
   */
  public static final class Performance {

    /**
     * Whether to optimize grid source sets, or not.
     *
     * <p><b>Default:</b> {@code true}
     */
    public final boolean optimizeGrids;

    /**
     * Whether to distribute grid sources near a site to smooth event rates.
     *
     * <p><b>Default:</b> {@code true}
     */
    public final boolean smoothGrids;

    /**
     * The partition or batch size to use when distributing
     * {@link SourceType#FAULT_SYSTEM} calculations.
     *
     * <p><b>Default:</b> {@code 1000}
     */
    public final int systemPartition;

    /**
     * The number of threads to use when distributing calculations.
     *
     * <p><b>Default:</b> {@link ThreadCount#ALL}
     */
    public final ThreadCount threadCount;

    /**
     * Whether to apply updated distance filter.
     *
     * <p><b>Default:</b> false
     */
    public final boolean distanceFilterUpdate;

    // note: distanceFilterUpdate is picked up when model loads
    // and cannot be changed with user supplied config

    private Performance(Builder b) {
      this.optimizeGrids = b.optimizeGrids;
      this.smoothGrids = b.smoothGrids;
      this.systemPartition = b.systemPartition;
      this.threadCount = b.threadCount;
      this.distanceFilterUpdate = b.distanceFilterUpdate;
    }

    private static final class Builder {

      Boolean optimizeGrids;
      Boolean smoothGrids;
      Integer systemPartition;
      ThreadCount threadCount;
      Boolean distanceFilterUpdate;

      Performance build() {
        checkNotNull(optimizeGrids);
        checkNotNull(smoothGrids);
        checkNotNull(systemPartition);
        checkNotNull(threadCount);
        checkNotNull(distanceFilterUpdate);
        return new Performance(this);
      }

      void copy(Performance that) {
        this.optimizeGrids = that.optimizeGrids;
        this.smoothGrids = that.smoothGrids;
        this.systemPartition = that.systemPartition;
        this.threadCount = that.threadCount;
        this.distanceFilterUpdate = that.distanceFilterUpdate;
      }

      void extend(Builder that) {
        if (that.optimizeGrids != null) {
          this.optimizeGrids = that.optimizeGrids;
        }
        if (that.smoothGrids != null) {
          this.smoothGrids = that.smoothGrids;
        }
        if (that.systemPartition != null) {
          this.systemPartition = that.systemPartition;
        }
        if (that.threadCount != null) {
          this.threadCount = that.threadCount;
        }
        if (that.distanceFilterUpdate != null) {
          this.distanceFilterUpdate = that.distanceFilterUpdate;
        }
      }
    }
  }

  @Override
  public String toString() {
    return GSON.toJson(this);
  }

  private static final Gson GSON = new GsonBuilder()
      .setPrettyPrinting()
      .enableComplexMapKeySerialization()
      .serializeNulls()
      .registerTypeHierarchyAdapter(Path.class, new PathConverter())
      .create();

  private static class PathConverter implements JsonSerializer<Path>, JsonDeserializer<Path> {

    @Override
    public Path deserialize(
        JsonElement json,
        Type type,
        JsonDeserializationContext context) throws JsonParseException {
      return Paths.get(json.getAsString());
    }

    @Override
    public JsonElement serialize(
        Path path,
        Type type,
        JsonSerializationContext context) {
      return new JsonPrimitive(path.toAbsolutePath().normalize().toString());
    }
  }

  /**
   * Save this config in JSON format to the speciifed directory.
   *
   * @param dir the directory to write to
   * @throws IOException if there is a problem writing the file
   */
  public void write(Path dir) throws IOException {
    Path file = dir.resolve(FILE_NAME);
    Writer writer = Files.newBufferedWriter(file, StandardCharsets.UTF_8);
    GSON.toJson(this, writer);
    writer.close();
  }

  /**
   * A builder of configuration instances.
   */
  public static final class Builder {

    private Hazard.Builder hazard;
    private Gmm.Builder gmm;
    private Disagg.Builder disagg;
    private Rate.Builder rate;
    private Output.Builder output;
    private Performance.Builder performance;

    private Builder() {
      hazard = new Hazard.Builder();
      gmm = new Gmm.Builder();
      disagg = new Disagg.Builder();
      rate = new Rate.Builder();
      output = new Output.Builder();
      performance = new Performance.Builder();
    }

    /**
     * Extend {@code this} builder to match {@code that} builder. Fields in
     * {@code that} builder take precedence unless they are not set.
     */
    public Builder extend(Builder that) {
      checkNotNull(that);
      this.hazard.extend(that.hazard);
      this.gmm.extend(that.gmm);
      this.disagg.extend(that.disagg);
      this.rate.extend(that.rate);
      this.output.extend(that.output);
      this.performance.extend(that.performance);
      return this;
    }

    /*
     * Those values for which web services require custom configurations
     * (overrides) are exposed below.
     */

    /**
     * Set the IMTs for which results should be calculated.
     *
     * @see gov.usgs.earthquake.nshmp.calc.CalcConfig.Hazard#imts
     */
    public Builder imts(Set<Imt> imts) {
      this.hazard.imts = checkNotNull(imts);
      return this;
    }

    /**
     * Set the data types for which results should be calculated.
     *
     * @see gov.usgs.earthquake.nshmp.calc.CalcConfig.Hazard#imts
     */
    public Builder dataTypes(Set<DataType> dataTypes) {
      this.output.dataTypes = checkNotNull(dataTypes);
      return this;
    }

    /**
     * Set the timespan for earthquake probabilities. Calling this method also
     * sets {@link Rate#valueFormat} to {@link ValueFormat#POISSON_PROBABILITY}
     * to ensure consistency.
     *
     * @see Rate#timespan
     */
    public Builder timespan(double timespan) {
      this.rate.timespan = timespan;
      this.rate.valueFormat = ValueFormat.POISSON_PROBABILITY;
      return this;
    }

    /**
     * Set the cutoff distance within which to include all sources when
     * computing earthquake rates or probabilities.
     *
     * @see Rate#distance
     */
    public Builder distance(double distance) {
      this.rate.distance = distance;
      return this;
    }

    /**
     * Build a new calculation configuration.
     */
    public CalcConfig build() {
      return new CalcConfig(this);
    }
  }
}
