package gov.usgs.earthquake.nshmp.gmm;

import static gov.usgs.earthquake.nshmp.gmm.Gmm.BA_08_BASE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.BASE_10_TO_E;
import static gov.usgs.earthquake.nshmp.gmm.GmmUtils.LN_G_CM_TO_M;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGA;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P01;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P02;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P03;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P05;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P06;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P075;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P08;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P12;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P15;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P16;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P6;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P75;
import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P8;
import static java.lang.Math.log10;
import static java.lang.Math.sqrt;

import java.util.Map;
import java.util.Optional;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

// spotless:off
// import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P1;
// import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P2;
// import static gov.usgs.earthquake.nshmp.gmm.Imt.SA0P5;
// import static gov.usgs.earthquake.nshmp.gmm.Imt.SA1P0;
// spotless:on

/**
 * Implementation of the ground motion relations for Puerto Rico by Motazedian
 * and Atkinson (2005) in order to evaluate the model for use in the 2025 Puerto
 * Rico and U.S. Virgin Islands NSHM.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation notes:</b><ul>
 *
 * <li>Published coefficients are defined in frequency-space. Frequencies are
 * converted to the closest existing spectral period in IMT.</li>
 *
 * <li>Support for spectral period 0.01s is provided using the same coefficients
 * as PGA.</li>
 *
 * <li>Support for spectral periods 0.02s, 0.03s, 0.075s, 0.15s, 0.25s, and 1.5s
 * is provided via interpolation of ground motion and sigma of adjacent periods
 * for which there are coefficients.</li>
 *
 * <li>Uses site-scaling model of BA08.</li>
 *
 * <p><b>Reference:</b> Motazedian, D. and Atkinson, G., 2005, Ground-motion
 * relations for Puerto Rico, in Mann, P., ed., Active tectonics and seismic
 * hazards of Puerto Rico, the Virgin Islands, and offshore areas: Geological
 * Society of America Special Paper 385, p. 61-80,
 * https://doi.org/10.1130/SPE385.
 *
 * <p><b>doi: </b> <a href="https://doi.org/10.1130/SPE385">10.1130/SPE385</a>
 *
 * <p><b>Component:</b> Geometric mean of two horizontal components
 *
 * @author U.S. Geological Survey
 * @see Gmm#MA_05
 */
public class MotazedianAtkinson_2005 implements GroundMotionModel {

  static final String NAME = "Motazedian & Atkinson (2005) BETA";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 9.5))
      .set(RRUP, Range.closed(0.0, 1000.0))
      .set(VS30, Range.singleton(760.0))
      .build();

  /*
   * Coefficients for MPRS IMT, consistent with BA08. Published PGA coefficients
   * are duplicated for 0.01 sec.
   */
  static final CoefficientContainer COEFFS_MPRS =
      new CoefficientContainer("MotazedianAtkinson05_mprs.csv");
  /*
   * Coefficients for use with InterpolatedGmm. Published coefficients are given
   * for frequency (Hz). These are converted to the closest available IMT
   * period. Published PGA coefficients are duplicated for 0.01 sec.
   */
  static final CoefficientContainer COEFFS_INTERP =
      new CoefficientContainer("MotazedianAtkinson05_interp.csv");

  /*
   * Interpolated IMTs added for MPRS Imts for which there are not published
   * coefficients.
   */
  private static final Map<Imt, Range<Imt>> INTERPOLATED_IMTS =
      Maps.immutableEnumMap(
          ImmutableMap.<Imt, Range<Imt>> builder()
              .put(SA0P02, Range.closed(SA0P01, SA0P06))
              .put(SA0P03, Range.closed(SA0P01, SA0P06))
              .put(SA0P05, Range.closed(SA0P01, SA0P06))
              .put(SA0P075, Range.closed(SA0P06, SA0P08))
              .put(SA0P15, Range.closed(SA0P12, SA0P16))
              .put(SA0P75, Range.closed(SA0P6, SA0P8))
              .build());
  // /* interpolate from MPRS IMTs instead of non-MPRS native IMTs */
  // private static final Map<Imt, Range<Imt>> INTERPOLATED_IMTS =
  // Maps.immutableEnumMap(
  // ImmutableMap.<Imt, Range<Imt>> builder()
  // .put(SA0P02, Range.closed(SA0P01, SA0P1))
  // .put(SA0P03, Range.closed(SA0P01, SA0P1))
  // .put(SA0P05, Range.closed(SA0P01, SA0P1))
  // .put(SA0P075, Range.closed(SA0P01, SA0P1))
  // .put(SA0P15, Range.closed(SA0P1, SA0P2))
  // .put(SA0P75, Range.closed(SA0P5, SA1P0))
  // .build());

  // distance centering coefficients
  private static final double C5 = -7.333;
  private static final double C6 = 2.333;
  // hinge function coefficients
  private static final double C7 = -1.8;
  private static final double C8 = 0.1;
  private static final double DIST_HINGE_LOWER = 75.0;
  private static final double DIST_HINGE_UPPER = 100.0;

  // convert log10 sigma to ln
  private static final double SIGMA = 0.28 * BASE_10_TO_E;

  private static final class Coefficients {

    final Imt imt;
    final double c1, c2, c3, c4;

    Coefficients(Imt imt, CoefficientContainer cc) {
      this.imt = imt;
      Map<String, Double> coeffs = cc.get(imt);
      c1 = coeffs.get("c1");
      c2 = coeffs.get("c2");
      c3 = coeffs.get("c3");
      c4 = coeffs.get("c4");
    }
  }

  private final Coefficients coeffs;
  // private final Coefficients coeffsPGA;
  //
  // /* Boore and Atkinson (2008) site scaling model */
  // private final BooreAtkinson_2008 siteAmp;

  // private final boolean interpolated;
  // private final GroundMotionModel interpolatedGmm;

  MotazedianAtkinson_2005(Imt imt) {
    this(imt, COEFFS_MPRS);
    // coeffs = new Coefficients(imt, COEFFS_MPRS);

    // for (Imt i : COEFFS_MPRS.imts()) {
    // System.out.println(i);
    // }
    // System.out.println();
    // for (Imt i : COEFFS_INTERP.imts()) {
    // System.out.println(i);
    // }
    // System.out.println();
    // coeffsPGA = new Coefficients(PGA, COEFFS);
    // siteAmp = (BooreAtkinson_2008) BA_08_BASE.instance(imt);

    // interpolated = INTERPOLATED_IMTS.containsKey(imt);
    // interpolatedGmm = interpolated
    // ? new InterpolatedGmm(Gmm.MA_05_INTERP, imt, INTERPOLATED_IMTS.get(imt))
    // : null;
  }

  MotazedianAtkinson_2005(Imt imt, CoefficientContainer cc) {
    coeffs = new Coefficients(imt, cc);
  }

  @Override
  public Imt imt() {
    return coeffs.imt;
  }

  @Override
  public LogicTree<GroundMotion> calc(GmmInput in) {
    // if (interpolated) {
    // return interpolatedGmm.calc(in);
    // }

    return calc(coeffs, in);
  }

  private LogicTree<GroundMotion> calc(
      Coefficients c,
      GmmInput in) {
    double μ = calcMean(c, in.rRup, in.Mw);

    double σ = SIGMA;
    return GroundMotions.createTree(μ, σ);
  }

  private final double calcMean(
      Coefficients c,
      double rRup,
      double Mw) {
    /* pass rRup and Mw instead of `in` */
    /* Eq. 6, log10(PSA) in cm/s/s */

    double r = getR(rRup, Mw);

    double hingeFunction = getHingeFunction(r, Mw);

    double μ = c.c1 + c.c2 * (Mw - 6) + c.c3 * (Mw - 6) * (Mw - 6) + hingeFunction + c.c4 * r;
    // convert from base 10 to base 2 and from cm/s/s to g
    return μ * BASE_10_TO_E - LN_G_CM_TO_M;
  }

  private static final double getR(double rRup, double Mw) {
    // Equation and coefficients defined in Table 2 caption
    double delta = C5 + C6 * Mw;
    return sqrt(rRup * rRup + delta * delta);
  }

  private static final double getHingeFunction(double dist, double Mw) {
    // Equation and coefficients defined in Table 2 caption
    double mFac = C7 + C8 * Mw;
    if (dist <= DIST_HINGE_LOWER) {
      // R <= 75 km
      return mFac * log10(dist);
    } else if (dist <= DIST_HINGE_UPPER) {
      // 75 km < R <= 100 km
      return mFac * log10(DIST_HINGE_LOWER);
    } else {
      // R >= 100 km
      return mFac * log10(DIST_HINGE_LOWER) - 0.5 * log10(dist / DIST_HINGE_UPPER);
    }
  }

  /*
   * Usgs implementation requires site scaling using BA08 model.
   *
   * MA05 is frequency based so it's coefficients don't align cleanly with MPRS
   * IMTs, which requires the use of InterpolatedGmm. We want to interpolate
   * MA05 to MPRS using periods converted from frequencies rather than the
   * nearest available MPRS IMTs. Since BA08 coefficients do not include the
   * MA05 periods we want to use for interpolating MA05, we run into GMM
   * initialization errors due to unsupported IMTs.
   *
   * If IMT requires MA05 interpolation, do this using the native MA05
   * coefficients (between periods not supported by BA08), and then compute the
   * BA08 site factor at one of the MPRS IMTs supported by BA08.
   */
  static class Usgs extends MotazedianAtkinson_2005 {
    Usgs(Imt imt) {
      super(imt);
      // coeffs = new Coefficients(imt, COEFFS_MPRS);
      coeffsPGA = new Coefficients(PGA, COEFFS_MPRS);

      /*
       * ...if imt is in BA_08_BASE.imts(), get it, otherwise do OptionalEmpty
       */
      if (BooreAtkinson_2008.COEFFS.imts().contains(imt)) {
        siteAmp = Optional.of((BooreAtkinson_2008) BA_08_BASE.instance(imt));
      } else {
        // we shouldn't get here since COEFFS_MPRS only contains MPRS, which are
        // all supported by BA08
        siteAmp = Optional.empty();
      }

      interpolated = INTERPOLATED_IMTS.containsKey(imt);
      // interpolatedGmm = interpolated
      // ? new InterpolatedGmm(Gmm.MA_05_INTERP, imt,
      // INTERPOLATED_IMTS.get(imt))
      // : null;

    }

    // private final Coefficients coeffs;
    private final Coefficients coeffsPGA;

    /* Boore and Atkinson (2008) site scaling model */
    // private final BooreAtkinson_2008 siteAmp;
    private final Optional<BooreAtkinson_2008> siteAmp;

    private final boolean interpolated;
    // private final GroundMotionModel interpolatedGmm;

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {

      // System.out.printf("Imt: %s GMM mprs\n", super.coeffs.imt);
      double μRef;
      if (interpolated) {
        // μRef = interpolatedGmm.calc(in).get(0).value().mean();
        // μRef =
        // Gmm.MA_05_BASE.instance(super.coeffs.imt).calc(in).get(0).value().mean();
        μRef = Gmm.MA_05_INTERP.instance(super.coeffs.imt).calc(in).get(0).value().mean();
      } else {
        μRef = super.calcMean(super.coeffs, in.rRup, in.Mw);
      }

      double μPga = super.calcMean(coeffsPGA, in.rRup, in.Mw);
      double site;
      if (siteAmp.isPresent()) {
        site = siteAmp.get().siteAmp(μPga, in.vs30);
      } else {
        site = 0;
      }
      double μAm = μRef + site;

      double σ = SIGMA;
      return GroundMotions.createTree(μAm, σ);

    }
  }

  static class Interp extends MotazedianAtkinson_2005 {
    static final String NAME = MotazedianAtkinson_2005.NAME + " : INTERPOLATED";

    Interp(Imt imt) {
      /* initialize coeffs with interpolation coefficient file */
      super(imt, COEFFS_INTERP);

      interpolated = INTERPOLATED_IMTS.containsKey(imt);
      interpolatedGmm = interpolated
          ? new InterpolatedGmm(Gmm.MA_05_INTERP, imt, INTERPOLATED_IMTS.get(imt))
          : null;
    }

    private final boolean interpolated;
    private final GroundMotionModel interpolatedGmm;

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      // assert interpolated;
      // System.out.printf("Imt: %s, GMM interpolated\n", super.coeffs.imt);
      if (interpolated) {
        return interpolatedGmm.calc(in);
      }
      return super.calc(super.coeffs, in);
    }
  }

}
