package gov.usgs.earthquake.nshmp.gmm;

import static com.google.common.io.Resources.getResource;
import static com.google.common.io.Resources.readLines;
import static gov.usgs.earthquake.nshmp.data.DoubleData.checkWeights;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.TABLE_DIR;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotions.EPI_WTS;
import static gov.usgs.earthquake.nshmp.gmm.GroundMotions.Z_SCORE_5_95;
import static java.lang.Math.exp;
import static java.lang.Math.log;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;
import static java.util.stream.Collectors.toUnmodifiableMap;

import java.io.IOException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.Text.Delimiter;
import gov.usgs.earthquake.nshmp.data.Interpolator;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionTables.GroundMotionTable.Position;
import gov.usgs.earthquake.nshmp.tree.LogicTree;

/**
 * Implementation of the PEER NGA-East ground motion model (Goulet et al.,
 * 2018). It is a composite model that consists of 17 median ground motion
 * models with period dependent weights. This class also includes the
 * implementations required for the USGS seed model logic tree.
 *
 * <p>In addition to the implementations of the multiple 'seed' models that were
 * used to generate (via Sammons mapping) the NGA-East component models, this
 * class includes updates to some of those models that are required for
 * different versions of the USGS seed GMM logic tree. Ground motions for most
 * models (seed and sammons) are computed via table lookups but several seed
 * updates are based on functional forms and implemented in separate classes.
 * For example, see {@link ShahjoueiPezeshk_2016} or {@link PezeshkEtAl_2018}.
 *
 * <p>On its own, NGA-East is a hard rock model returning results for a site
 * class where Vs30 = 3000 m/s. To accomodate other site classes, the Stewart et
 * al. (2017) CEUS site amplification model is used. This model is applicable to
 * 200 ≤ Vs30 ≤ 2000 m/s. For vs30 &lt; 200 the model reasonably extrapolates
 * down to 150 m/s (@pmpowers J. Stewart, pers. comm., 1/26/18). For Vs30 &gt;
 * 2000 m/s, the model scales linearly to zero at Vs30 = 3000 m/s.
 *
 * <p> The aleatory variability model supplied by instances of NGA-East and the
 * USGS seed tree consists of a logic tree of the EPRI (2013) model and model
 * developed by a panel convened by the NSHMP that includes site-to-site
 * variability.
 *
 * <p><b>Implementation history:</b> The initial implementation of NGA-East was
 * based on a version of the model developed specifically for USGS applications
 * (Goulet et al., 2017), which turned out to be identical to the final NGA-East
 * model developed for the NRC. As part of the 2023 NSHM update the Hashash et
 * al. (2020) non-linear adjustments were included in the site amplification
 * model, and GMM variants were added to support coastal plain adjustments and
 * direct adjustments to the NGA-East hard rock model. Both adjustments taper
 * in/out as a function of zSed. The application of the site amplification term
 * from 2018 was also updated to better represent the 5th, 50th, and 95th
 * percentiles of ground motion. Note, however, that the site term is still
 * applied to the median ground in ground motion space, not rate space as it
 * would be if the site amplification branches were carried all the way through
 * to hazard.
 *
 * <p>Inclusion of updated seed models by Boore (2020) and Pezeshk et al. (2018)
 * were considered for the 2023 update, but they were found to make a negligable
 * difference in hazard and ultimately not included. In the names of the seed
 * models, a (ᵀ) indicates that the model is part of the USGS seed model logic
 * tree; a (*) indicates that the model is an update to a seed model that was
 * used to develop the NGA-East Sammon's mapped models.
 *
 * <p><b>Note:</b> Direct instantiation of {@code GroundMotionModel}s is
 * prohibited. Use {@link Gmm#instance(Imt)} to retrieve an instance for a
 * desired {@link Imt}.
 *
 * <p><b>Implementation note:</b> Mean values are NOT clamped per
 * {@link GmmUtils#ceusMeanClip(Imt, double)}.
 *
 * <p><b>Reference:</b> Goulet, C., Bozorgnia, Y., Kuehn, N., Al Atik, L.,
 * Youngs, R., Graves, R., Atkinson, G., 2017, NGA-East ground-motion models for
 * the U.S. Geological Survey national seismic hazard maps: PEER Report No.
 * 2017/03, 180 p.
 *
 * <p><b>Reference:</b> Goulet CA, Bozorgnia Y, Abrahamson N, et al. (2018)
 * Central and eastern North America ground— motion characterization—NGA-East
 * final report. Report no. 2018/08. Berkeley, CA: PEER, 817 pp.
 *
 * <p><b>Reference:</b> Stewart, J., Parker, G., Atkinson, G., Boore, D., and
 * Hashash, Y., and Silva, W., 2020, Ergodic site amplification model for
 * central and eastern North America: Earthquake Spectra 36(1): 42–68.
 *
 * <p><b>Reference:</b> Hashash, Y., Harmon, J., Ilhan, O., Parker, G., and
 * Stewart, 2020, Nonlinear site amplification model for ergodic seismic hazard
 * analysis in central and eastern North America. Earthquake Spectra 36(1):
 * 69–86.
 *
 * <p><b>Reference:</b>Stewart, J., Parker, G., Al Atik, L, et al. (2019)
 * Site-to-site standard deviation model for central and eastern North America.
 * UC E-Scholarship publication. Available at: <a
 * href="https://escholarship.org/uc/item/2sc5g220">
 * https://escholarship.org/uc/item/2sc5g220</a> (accessed 18 November 2019).
 *
 * <p><b>Reference:</b> Chapman, M.C. and Guo, Z., 2021, A response spectral
 * ratio model to account for amplification and attenuation effects in the
 * Atlantic and Gulf coastal plain: Bulletin of the Seismological Society of
 * America, v. 111, n. 4, p. 1849-1867, doi: <a
 * href="https://doi.org/10.1785/0120200322"> 10.1785/0120200322</a>.
 *
 * <p><b>Component:</b> average horizontal (RotD50)
 *
 * @author U.S. Geological Survey
 * @see Gmm#NGA_EAST_2018
 * @see Gmm#NGA_EAST_SEEDS_2018
 */
public abstract class NgaEast implements GroundMotionModel {

  /*
   * Developer notes:
   *
   * Note: supported periods are derived from sigma coefficient files. Several
   * supported periods have been commented out because they are not represented
   * in the site amplification model.
   *
   * When supplied with tables for the 17 usgs models, 0.01s is present with
   * values distinct from PGA, however the seed models are missing this period.
   * We therefore duplicate PGA for 0.01s ground motion values in the seed model
   * tables. Because a coefficient table (sigma coeffs in this case) is
   * referenced for supported IMTs, the ground motion tables across seed and
   * sammons models must be consistent.
   *
   * Missing PGV tables: Grazier, PEER_EX, PEER_GP, PZCT15_M1SS, PZCT15_M2ES
   *
   * Notes on sigma:
   *
   * NGA-East for USGS (2017) recommended use of an updated EPRI (2013) model
   * (Table 5.5), which is currently considered as one independent branch of an
   * aleatory variability logic tree.
   *
   * The NSHMP asked PEER panel to develop an improved φ_s2s model. This
   * necessitated including τ and φ_ss as independent terms. Only the 'global' τ
   * and φ_ss models are considered, per Linda Al Atik's recommendation. Because
   * no statistical uncertianty model was developed for φ_s2s, we only consider
   * the 'central' branch for τ and φ_ss terms. Sensitivity studies show that
   * consideration of only the 'global' and 'central' branches of τ and φ_ss has
   * little effect on the final σ. This 'panel' model is the second branch in
   * the aleatory variability logic tree.
   */

  static final String NAME = "NGA-East";

  static final Constraints CONSTRAINTS = Constraints.builder()
      .set(MW, Range.closed(4.0, 8.2))
      .set(RJB, Range.closed(0.0, 1500.0))
      .set(VS30, Range.closed(200.0, 3000.0))
      .build();

  /*
   * Sigma coefficients for global model from tables 5-1 (tau) and 5-2 (phi) and
   * EPRI model from table 5-5.
   */
  static CoefficientContainer COEFFS_SIGMA_PANEL;
  static CoefficientContainer COEFFS_SIGMA_EPRI;

  /* NGA-East short period adjustments for 2023 NSHM. */
  static CoefficientContainer COEFFS_NGA_ADJ_2023;

  /* Immutable, ordered map of seed model weights. */
  static Map<String, Double> USGS_SEED_WEIGHTS;

  static {
    COEFFS_SIGMA_PANEL = new CoefficientContainer("nga-east-usgs-sigma-panel.csv");
    COEFFS_SIGMA_EPRI = new CoefficientContainer("nga-east-usgs-sigma-epri.csv");
    COEFFS_NGA_ADJ_2023 = new CoefficientContainer("nga-east-usgs-adj-2023.csv");
    USGS_SEED_WEIGHTS = initSeedWeights();
  }

  private static Map<String, Double> initSeedWeights() {
    try {
      URL wtsUrl = getResource(TABLE_DIR + "nga-east-seed-weights.dat");
      Map<String, Double> wtMap = readLines(wtsUrl, StandardCharsets.UTF_8)
          .stream()
          .skip(1)
          .map(line -> Text.splitToList(line, Delimiter.COMMA))
          .collect(toUnmodifiableMap(
              e -> e.get(0),
              e -> Double.valueOf(e.get(1))));
      checkWeights(wtMap.values());
      return wtMap;

    } catch (IOException ioe) {
      throw new RuntimeException(ioe);
    }
  }

  /* Panel and EPRI sigma models and wieghts */
  private static final String[] SIGMA_IDS = { "σ-epri", "σ-usgs" };
  private static final double[] SIGMA_WTS = { 0.8, 0.2 };

  /* φ_s2s constants */
  private static final double VΦ1 = 1200.0;
  private static final double VΦ2 = 1500.0;

  private static final class CoefficientsSigmaPanel {

    /* τ coefficients, global model, central branch */
    final double τ1, τ2, τ3, τ4;

    /* φ_ss coefficients, global model, central branch */
    final double a, b;

    /* φ_s2s coefficients */
    final double φs2s1, φs2s2;

    CoefficientsSigmaPanel(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      τ1 = coeffs.get("t1");
      τ2 = coeffs.get("t2");
      τ3 = coeffs.get("t3");
      τ4 = coeffs.get("t4");
      a = coeffs.get("ss_a");
      b = coeffs.get("ss_b");
      φs2s1 = coeffs.get("s2s1");
      φs2s2 = coeffs.get("s2s2");
    }
  }

  /* Updated EPRI (2013); initial NGAE-East team recommendation. */
  private static final class CoefficientsSigmaEpri {

    final double τ_m5, φ_m5, τ_m6, φ_m6, τ_m7, φ_m7;

    CoefficientsSigmaEpri(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      τ_m5 = coeffs.get("tau_M5");
      φ_m5 = coeffs.get("phi_M5");
      τ_m6 = coeffs.get("tau_M6");
      φ_m6 = coeffs.get("phi_M6");
      τ_m7 = coeffs.get("tau_M7");
      φ_m7 = coeffs.get("phi_M7");
    }
  }

  /* NGA-East short period adjustments for 2023 NSHM. */
  private static final class CoefficientsNgaAdj2023 {

    final double nga_adj;

    CoefficientsNgaAdj2023(Imt imt, CoefficientContainer cc) {
      Map<String, Double> coeffs = cc.get(imt);
      nga_adj = coeffs.get("nga_adj");
    }
  }

  private final CoefficientsSigmaPanel σCoeffsPanel;
  private final CoefficientsSigmaEpri σCoeffsEpri;
  final CoefficientsNgaAdj2023 coeffsNgaAdj;

  final Imt imt;

  NgaEast(Imt imt) {
    σCoeffsPanel = new CoefficientsSigmaPanel(imt, COEFFS_SIGMA_PANEL);
    σCoeffsEpri = new CoefficientsSigmaEpri(imt, COEFFS_SIGMA_EPRI);
    coeffsNgaAdj = new CoefficientsNgaAdj2023(imt, COEFFS_NGA_ADJ_2023);
    this.imt = imt;
  }

  @Override
  public Imt imt() {
    return imt;
  }

  /* Apply zSed-dependent coastal plain adjustment . */
  boolean cpa() {
    return false;
  }

  /* Apply zSed-dependent NGA-East adjustments. */
  boolean adjusted() {
    return false;
  }

  double sigmaEpri(double Mw) {
    return sigmaEpri(σCoeffsEpri, Mw);
  }

  double sigmaPanel(double Mw, double vs30) {
    return sigmaPanel(σCoeffsPanel, Mw, vs30);
  }

  /* σ branch 1: Updated EPRI (2013); initial recomendation. */
  private static double sigmaEpri(CoefficientsSigmaEpri c, double Mw) {
    double τ;
    double φ;
    if (Mw <= 5.0) {
      τ = c.τ_m5;
      φ = c.φ_m5;
    } else if (Mw <= 6.0) {
      τ = Interpolator.findY(5.0, c.τ_m5, 6.0, c.τ_m6, Mw);
      φ = Interpolator.findY(5.0, c.φ_m5, 6.0, c.φ_m6, Mw);
    } else if (Mw <= 7.0) {
      τ = Interpolator.findY(6.0, c.τ_m6, 7.0, c.τ_m7, Mw);
      φ = Interpolator.findY(6.0, c.φ_m6, 7.0, c.φ_m7, Mw);
    } else {
      τ = c.τ_m7;
      φ = c.φ_m7;
    }
    return Maths.hypot(φ, τ);
  }

  /* σ branch 2: Panel recommendation with φ_s2s. */
  private static double sigmaPanel(
      CoefficientsSigmaPanel c,
      double Mw,
      double vs30) {

    /* τ model; global branch only; Equation 5-1 */
    double τ = tau(Mw, c.τ1, c.τ2, c.τ3, c.τ4);

    /* φ_ss model; global branch only; Equation 5-2 */
    double φ_ss = phi_ss(Mw, c.a, c.b);

    /* φ_s2s model; single branch; Stewart et al. (2019) */
    double φ_s2s = phi_s2s(vs30, c.φs2s1, c.φs2s2);

    return Maths.hypot(τ, φ_ss, φ_s2s);
  }

  /* τ: Equation 5.1 */
  private static double tau(
      double Mw,
      double τ1,
      double τ2,
      double τ3,
      double τ4) {

    if (Mw <= 4.5) {
      return τ1;
    } else if (Mw <= 5.0) {
      return τ1 + (τ2 - τ1) * (Mw - 4.5) / 0.5;
    } else if (Mw <= 5.5) {
      return τ2 + (τ3 - τ2) * (Mw - 5.0) / 0.5;
    } else if (Mw <= 6.5) {
      return τ3 + (τ4 - τ3) * (Mw - 5.5);
    }
    return τ4;
  }

  /* φ_ss: Equation 5.2 */
  private static double phi_ss(
      double Mw,
      double a,
      double b) {

    if (Mw <= 5.0) {
      return a;
    } else if (Mw <= 6.5) {
      return a + (Mw - 5.0) * (b - a) / 1.5;
    }
    return b;
  }

  /* φ_s2s: Stewart et al. */
  private static double phi_s2s(
      double vs30,
      double φs2s1,
      double φs2s2) {

    if (vs30 < VΦ1) {
      return φs2s1;
    } else if (vs30 < VΦ2) {
      return φs2s1 - ((φs2s1 - φs2s2) / (VΦ2 - VΦ1)) * (vs30 - VΦ1);
    }
    return φs2s2;
  }

  /* Implementation of sammons map models. */
  static class NgaEast_2018 extends NgaEast {

    static final String NAME = NgaEast.NAME + " (2018)";

    static final int MODEL_COUNT = 17;
    static final String[] MEAN_IDS = IntStream.range(1, MODEL_COUNT + 1)
        .mapToObj(i -> "Sammons" + i)
        .collect(Collectors.toList())
        .toArray(new String[MODEL_COUNT]);

    final GroundMotionTable[] tables;
    final GroundMotionTable[] pgaTables;
    final SiteAmp_2018 siteAmp;
    final double[] μWts;

    NgaEast_2018(Imt imt) {
      this(imt, new SiteAmp_2018(imt));
    }

    NgaEast_2018(Imt imt, SiteAmp_2018 siteAmp) {
      super(imt);
      this.tables = GroundMotionTables.getNgaEast(imt);
      this.pgaTables = GroundMotionTables.getNgaEast(Imt.PGA);
      this.siteAmp = siteAmp;
      this.μWts = GroundMotionTables.getNgaEastWeights(imt);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      Position p = tables[0].position(in.rRup, in.Mw);
      double[] μs = new double[MODEL_COUNT];
      for (int i = 0; i < MODEL_COUNT; i++) {
        double μ = tables[i].get(p);
        double μPga = exp(pgaTables[i].get(p));
        SiteTerm fSite = siteAmp.calc(μPga, in.vs30);
        μs[i] = fSite.apply(μ);
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, in.vs30) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, μWts,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  public static void main(String[] args) {
    // GmmInput in1 = GmmInput.builder().withDefaults().build();
    GmmInput in1 = GmmInput.builder().withDefaults().zSed(4.0).build();

    Imt imt = Imt.SA0P05;

    GroundMotionModel gmm18 = Gmm.NGA_EAST_2018.instance(imt);
    GroundMotionModel gmm23 = Gmm.NGA_EAST_2023_ADJUSTED.instance(imt);

    LogicTree<GroundMotion> gmm18lt = gmm18.calc(in1);
    LogicTree<GroundMotion> gmm23lt = gmm23.calc(in1);
    System.out.println(gmm18lt);
    System.out.println(gmm23lt);

    GroundMotion gm18 = GroundMotions.combine(gmm18lt);
    GroundMotion gm23 = GroundMotions.combine(gmm23lt);

    System.out.println("2018: " + gm18);
    System.out.println("2023: " + gm23);

    // GroundMotion gm1 = GroundMotions.combine(gmm.calc(in1));
    // GroundMotion gm2 = GroundMotions.combine(gmm.calc(in2));
    // System.out.println(gm1);
    // System.out.println(gm2);
  }

  /*
   * Updated NGA-East for use in the 2023 nshm-conus update. This model includes
   * (1) the final published nonlinear site amplification model of Hshash et al.
   * (2020) and (2) the Gulf and Atlantic coastal plain effects model of Chapman
   * & Guo (2021).
   */
  static class NgaEast_2023 extends NgaEast_2018 {

    static final String NAME = "NGA-East (2023)";

    NgaEast_2023(Imt imt) {
      super(imt, new SiteAmp_2023(imt));
    }

    /*
     * Developer notes: The current implementation assumes a reference site
     * condition of Vs30 = 1000 m/s. This means that we apply the full Chapman &
     * Guo PSA ratio if Vs30 = 1000 and scale it by the ratio of the site term
     * at the target Vs30 to the reference Vs30.
     */
    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      double zScale = zSiteScale(in.zSed);
      double μAdj = adjusted()
          ? (1.0 - zScale) * coeffsNgaAdj.nga_adj
          : 0.0;
      boolean cpa = cpa() && zScale > 0.0;
      double fCpa = cpa
          ? log(ChapmanGuo_2021.cpaPsaRatio(imt, in.zSed, in.Mw, in.rJB))
          : 0.0;
      Position p = tables[0].position(in.rRup, in.Mw);
      double[] μs = new double[MODEL_COUNT];
      for (int i = 0; i < MODEL_COUNT; i++) {
        double μ = tables[i].get(p) + μAdj;
        double μPga = exp(pgaTables[i].get(p) + μAdj);
        SiteTerm fSite = siteAmp.calc(μPga, in.vs30);
        double cpaTerm = 0.0;
        if (cpa) {
          double cpaSiteCorr = siteAmp.calc(μPga, ChapmanGuo_2021.VS_REF).fs;
          cpaTerm = fCpa - cpaSiteCorr * zScale;
        }
        μs[i] = fSite.apply(μ) + cpaTerm;
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, in.vs30) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, μWts,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  static class NgaEast_2023_Adjusted extends NgaEast_2023 {
    static final String NAME = NgaEast.NAME + " (2023, adjusted)";

    NgaEast_2023_Adjusted(Imt imt) {
      super(imt);
    }

    @Override
    boolean adjusted() {
      return true;
    }
  }

  static class NgaEast_2023_Cpa extends NgaEast_2023 {
    static final String NAME = NgaEast.NAME + " (2023, CPA)";

    NgaEast_2023_Cpa(Imt imt) {
      super(imt);
    }

    @Override
    boolean cpa() {
      return true;
    }
  }

  /*
   * Implementation of USGS Seed model logic tree. All models but SP16 are table
   * based; SP16 is added to the median ground motion array last. NOTE that the
   * model ignores the SP16 aleatory variability model for consistency with the
   * other seed models.
   *
   * PGV: rather than use averaged tables for seed that don't provide PGV tables
   * (inital implementation), we delegate to UsgsPgvSupport. This requires quick
   * short circuiting of those Gmms lacking support. In doing so we use the
   * individual seed models which results in repeated position lookups. Both
   * sigma models considered include coefficients for PGV.
   */
  static class UsgsSeeds_2018 extends NgaEast {
    static final String NAME = NgaEast.NAME + " Seed Tree (2018)";
    static final String SEED_PREFIX = "NGA_EAST_SEED_";
    static final String SP16_ID = "SP16";
    static final Set<Gmm> NO_PGV_SEEDS = EnumSet.of(
        Gmm.NGA_EAST_SEED_GRAIZER, // not used
        Gmm.NGA_EAST_SEED_GRAIZER16,
        Gmm.NGA_EAST_SEED_GRAIZER17,
        Gmm.NGA_EAST_SEED_PEER_EX, // not used
        Gmm.NGA_EAST_SEED_PEER_GP,
        Gmm.NGA_EAST_SEED_PZCT15_M1SS,
        Gmm.NGA_EAST_SEED_PZCT15_M2ES);
    static final Set<String> NO_PGV_IDS = Set.of(
        "Graizer",
        "Graizer16",
        "Graizer17",
        "PEER_EX",
        "PEER_GP",
        "PZCT15_M1SS",
        "PZCT15_M2ES");

    /* Valid Seed IDs for table based models; skips SP16 */
    static final List<String> SEED_TABLE_IDS; // size = 13
    static final List<String> SEED_TABLE_IDS_PGV; // size = 8

    /* All Gmms */
    static final List<Gmm> GMMS; // pgv helper, size = 14

    /* All ground motion mean model IDs and weights */
    static final String[] MEAN_IDS;
    static final double[] MEAN_WTS;

    final Map<Gmm, GroundMotionTable> tables;
    final Map<Gmm, GroundMotionTable> pgaTables;
    final ShahjoueiPezeshk_2016 sp16;
    final ShahjoueiPezeshk_2016 sp16pga;
    final SiteAmp_2018 siteAmp;

    static {
      List<String> tableIds = new ArrayList<>(); // valid IDs for tables
      List<String> tableIdsPgv = new ArrayList<>(); // valid IDs for PGV tables
      List<Gmm> gmms = new ArrayList<>();
      List<String> μIds = new ArrayList<>();
      List<Double> μWts = new ArrayList<>();
      USGS_SEED_WEIGHTS.entrySet().forEach(e -> {
        String seedId = e.getKey();
        if (!seedId.equals(SP16_ID)) {
          tableIds.add(seedId);
          if (!NO_PGV_IDS.contains(seedId)) {
            tableIdsPgv.add(seedId);
          }
        }
        gmms.add(Gmm.valueOf(SEED_PREFIX + e.getKey().toUpperCase()));
        μIds.add("Seed (" + seedId + ")");
        μWts.add(e.getValue());
      });
      SEED_TABLE_IDS = List.copyOf(tableIds);
      SEED_TABLE_IDS_PGV = List.copyOf(tableIdsPgv);
      GMMS = List.copyOf(gmms);
      MEAN_IDS = μIds.toArray(new String[μIds.size()]);
      MEAN_WTS = μWts.stream()
          .mapToDouble(Double::doubleValue)
          .toArray();
    }

    UsgsSeeds_2018(Imt imt) {
      this(imt, new SiteAmp_2018(imt));
    }

    UsgsSeeds_2018(Imt imt, SiteAmp_2018 siteAmp) {
      super(imt);
      this.tables = getImtTables(imt);
      this.pgaTables = getImtTables(Imt.PGA);
      this.sp16 = (ShahjoueiPezeshk_2016) Gmm.NGA_EAST_SEED_SP16.instance(imt);
      this.sp16pga = (ShahjoueiPezeshk_2016) Gmm.NGA_EAST_SEED_SP16.instance(Imt.PGA);
      this.siteAmp = siteAmp;
    }

    private static Map<Gmm, GroundMotionTable> getImtTables(Imt imt) {
      List<String> ids = (imt == Imt.PGV) ? SEED_TABLE_IDS_PGV : SEED_TABLE_IDS;
      return GroundMotionTables.getNgaEastSeeds(ids, imt)
          .entrySet().stream()
          .collect(toUnmodifiableMap(
              e -> Gmm.valueOf(SEED_PREFIX + e.getKey().toUpperCase()),
              Entry::getValue));
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      Position p = tables.values().iterator().next().position(in.rRup, in.Mw);
      double[] μs = new double[GMMS.size()];
      for (int i = 0; i < GMMS.size(); i++) {
        Gmm seed = GMMS.get(i);
        if (seed == Gmm.NGA_EAST_SEED_SP16) {
          double μ = sp16.calcMeanRock(in.Mw, in.rJB);
          double μPga = exp(sp16pga.calcMeanRock(in.Mw, in.rJB));
          SiteTerm fSite = siteAmp.calc(μPga, in.vs30);
          μs[i] = fSite.apply(μ);
        } else if (imt == Imt.PGV && NO_PGV_SEEDS.contains(seed)) {
          // site is included when calling individual seed
          μs[i] = UsgsPgvSupport.calcAB20Pgv(seed, in).mean();
        } else {
          double μ = tables.get(seed).get(p);
          double μPga = exp(pgaTables.get(seed).get(p));
          SiteTerm fSite = siteAmp.calc(μPga, in.vs30);
          μs[i] = fSite.apply(μ);
        }
      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, in.vs30) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, MEAN_WTS,
          SIGMA_IDS, σs, SIGMA_WTS);
    }
  }

  /*
   * Updated NGA-East for use in the 2023 nshm-conus update. This model includes
   * (1) the final published nonlinear site amplification model of Hshash et al.
   * (2020) and (2) the Gulf and Atlantic coastal plain effects model of Chapman
   * & Guo (2021).
   */
  static class UsgsSeeds_2023 extends UsgsSeeds_2018 {

    static final String NAME = NgaEast.NAME + " Seed Tree (2023)";

    UsgsSeeds_2023(Imt imt) {
      super(imt, new SiteAmp_2023(imt));
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      GmmInput inRock = GmmInput.builder().fromCopy(in).vs30(3000).build();
      double zScale = zSiteScale(in.zSed);
      double μAdj = adjusted()
          ? (1.0 - zScale) * coeffsNgaAdj.nga_adj
          : 0.0;
      boolean cpa = cpa() && zScale > 0.0;
      double fCpa = cpa
          ? log(ChapmanGuo_2021.cpaPsaRatio(imt, in.zSed, in.Mw, in.rJB))
          : 0.0;
      Position p = tables.values().iterator().next().position(in.rRup, in.Mw);
      double[] μs = new double[GMMS.size()];
      for (int i = 0; i < GMMS.size(); i++) {
        Gmm seed = GMMS.get(i);

        double μ;
        double μPga;

        if (seed == Gmm.NGA_EAST_SEED_SP16) {
          μ = sp16.calcMeanRock(in.Mw, in.rJB);
          μPga = sp16pga.calcMeanRock(in.Mw, in.rJB);
        } else if (imt == Imt.PGV && NO_PGV_SEEDS.contains(seed)) {
          // note rock input; TODO this doesn't really work as the PGV
          // is conditioned on vs30 so site should already be applied
          μ = UsgsPgvSupport.calcAB20Pgv(seed, inRock).mean();
          μPga = pgaTables.get(seed).get(p);
        } else {
          μ = tables.get(seed).get(p);
          μPga = pgaTables.get(seed).get(p);
        }

        μ += μAdj;
        μPga = exp(μPga + μAdj);

        SiteTerm fSite = siteAmp.calc(μPga, in.vs30);
        double cpaTerm = 0.0;
        if (cpa) {
          double cpaSiteCorr = siteAmp.calc(μPga, ChapmanGuo_2021.VS_REF).fs;
          cpaTerm = fCpa - cpaSiteCorr * zScale;
        }
        μs[i] = fSite.apply(μ) + cpaTerm;

      }
      double[] σs = new double[] {
          sigmaEpri(in.Mw),
          sigmaPanel(in.Mw, in.vs30) };
      return GroundMotions.createTree(
          MEAN_IDS, μs, MEAN_WTS,
          SIGMA_IDS, σs, SIGMA_WTS);
    }

  }

  static class UsgsSeeds_2023_Adjusted extends UsgsSeeds_2023 {
    static final String NAME = NgaEast.NAME + " Seed Tree (2023, adjusted)";

    UsgsSeeds_2023_Adjusted(Imt imt) {
      super(imt);
    }

    @Override
    boolean adjusted() {
      return true;
    }
  }

  static class UsgsSeeds_2023_Cpa extends UsgsSeeds_2023 {
    static final String NAME = NgaEast.NAME + " Seed Tree (2023, CPA)";

    UsgsSeeds_2023_Cpa(Imt imt) {
      super(imt);
    }

    @Override
    boolean cpa() {
      return true;
    }
  }

  static abstract class Seed extends NgaEast {
    static final String NAME = NgaEast.NAME + " : Seed : ";

    final String id;
    final GroundMotionTable table;
    final GroundMotionTable pgaTable;
    final SiteAmp_2018 siteAmp;

    Seed(String id, Imt imt) {
      super(imt);
      this.id = id;
      this.table = GroundMotionTables.getNgaEastSeed(id, imt);
      this.pgaTable = GroundMotionTables.getNgaEastSeed(id, Imt.PGA);
      this.siteAmp = new SiteAmp_2018(imt);
    }

    @Override
    public LogicTree<GroundMotion> calc(GmmInput in) {
      double σEpri = sigmaEpri(in.Mw);
      double σPanel = sigmaPanel(in.Mw, in.vs30);
      // PGV fails if 'table' used; all seeds support PGA natively
      Position p = pgaTable.position(in.rRup, in.Mw);
      double μPga = exp(pgaTable.get(p));
      SiteTerm fSite = siteAmp.calc(μPga, in.vs30);

      Gmm seedGmm = Gmm.valueOf(UsgsSeeds_2018.SEED_PREFIX + id.toUpperCase());
      GmmInput inRock = GmmInput.builder().fromCopy(in).vs30(3000).build();
      double μ = (imt == Imt.PGV && UsgsSeeds_2018.NO_PGV_SEEDS.contains(seedGmm))
          ? UsgsPgvSupport.calcAB20Pgv(seedGmm, inRock).mean()
          : table.get(p);
      μ = fSite.apply(μ);

      /*
       * TODO the order of operations for PGV for individual seeds is consistent
       * with seed 2023 but not seed 2018 and needs to be rechecked. Conditional
       * PGV is dependent on vs30; should we realy be using the consitional
       * model with vs30=3000
       */
      double σ = Maths.srssWeighted(
          new double[] { σEpri, σPanel },
          SIGMA_WTS);
      return GroundMotions.createTree(μ, σ);
    }
  }

  static final class Seed_1CCSP extends Seed {
    static final String ID = "1CCSP";
    static final String NAME = Seed.NAME + ID;

    Seed_1CCSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_1CVSP extends Seed {
    static final String ID = "1CVSP";
    static final String NAME = Seed.NAME + ID;

    Seed_1CVSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_2CCSP extends Seed {
    static final String ID = "2CCSP";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_2CCSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_2CVSP extends Seed {
    static final String ID = "2CVSP";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_2CVSP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_a04 extends Seed {
    static final String ID = "B_a04";
    static final String NAME = Seed.NAME + ID;

    Seed_B_a04(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_ab14 extends Seed {
    static final String ID = "B_ab14";
    static final String NAME = Seed.NAME + ID;

    Seed_B_ab14(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_ab95 extends Seed {
    static final String ID = "B_ab95";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_B_ab95(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_bca10d extends Seed {
    static final String ID = "B_bca10d";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_B_bca10d(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_bs11 extends Seed {
    static final String ID = "B_bs11";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_B_bs11(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B_sgd02 extends Seed {
    static final String ID = "B_sgd02";
    static final String NAME = Seed.NAME + ID;

    Seed_B_sgd02(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B20_ab14mod1 extends Seed {
    static final String ID = "B20_ab14mod1";
    static final String NAME = Seed.NAME + ID + "*";

    Seed_B20_ab14mod1(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B20_ab14mod2 extends Seed {
    static final String ID = "B20_ab14mod2";
    static final String NAME = Seed.NAME + ID + "*";

    Seed_B20_ab14mod2(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_B20_bca10d extends Seed {
    static final String ID = "B20_bca10d";
    static final String NAME = Seed.NAME + ID + "*";

    Seed_B20_bca10d(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_Frankel extends Seed {
    static final String ID = "Frankel";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_Frankel(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_Graizer extends Seed {
    static final String ID = "Graizer";
    static final String NAME = Seed.NAME + ID;

    Seed_Graizer(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_Graizer16 extends Seed {
    static final String ID = "Graizer16";
    static final String NAME = Seed.NAME + ID + "ᵀ*";

    Seed_Graizer16(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_Graizer17 extends Seed {
    static final String ID = "Graizer17";
    static final String NAME = Seed.NAME + ID + "ᵀ*";

    Seed_Graizer17(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_HA15 extends Seed {
    static final String ID = "HA15";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_HA15(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PEER_EX extends Seed {
    static final String ID = "PEER_EX";
    static final String NAME = Seed.NAME + ID;

    Seed_PEER_EX(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PEER_GP extends Seed {
    static final String ID = "PEER_GP";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_PEER_GP(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PZCT15_M1SS extends Seed {
    static final String ID = "PZCT15_M1SS";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_PZCT15_M1SS(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_PZCT15_M2ES extends Seed {
    static final String ID = "PZCT15_M2ES";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_PZCT15_M2ES(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_SP15 extends Seed {
    static final String ID = "SP15";
    static final String NAME = Seed.NAME + ID;

    Seed_SP15(Imt imt) {
      super(ID, imt);
    }
  }

  static final class Seed_YA15 extends Seed {
    static final String ID = "YA15";
    static final String NAME = Seed.NAME + ID + "ᵀ";

    Seed_YA15(Imt imt) {
      super(ID, imt);
    }
  }

  /*
   * Stewart et al. site amplification model.
   *
   * The model is applicable to 200 ≤ vs30 ≤ 2000 m/s and smoothly scales from
   * 200 m/s up to 3000 m;s and reasonably extrapolates down to 150 m/s.
   */
  static class SiteAmp_2018 {

    private static final CoefficientContainer COEFFS = new CoefficientContainer(
        "nga-east-usgs-siteamp.csv");

    private static final double V_MIN = 150.0;
    private static final double V_MAX = 3000.0;
    private static final double V_LIN_REF = 760.0;
    private static final double VL = 200.0;
    private static final double VU = 2000.0;

    private static final double VW1 = 600.0;
    private static final double VW2 = 400.0;
    private static final double WT1 = 0.767; // impedance model @ VW1
    private static final double WT2 = 0.1; // impedance model @ VW2
    private static final double WT_SCALE = (WT1 - WT2) / (log(VW1) - log(VW2)); // ≈1.65

    final Coefficients c;
    final double zScore;

    private static final class Coefficients {

      final Imt imt;
      final double c, v1, v2, vf, σvc, σl, σu;
      final double f760i, f760g, f760iσ, f760gσ;
      final double f3, f4, f4mod, f5, vc, σc;

      Coefficients(Imt imt, CoefficientContainer cc) {
        this.imt = imt;
        Map<String, Double> coeffs = cc.get(imt);
        c = coeffs.get("c");
        v1 = coeffs.get("V1");
        v2 = coeffs.get("V2");
        vf = coeffs.get("Vf");
        σvc = coeffs.get("sig_vc");
        σl = coeffs.get("sig_l");
        σu = coeffs.get("sig_u");
        f760i = coeffs.get("f760i");
        f760iσ = coeffs.get("f760is");
        f760g = coeffs.get("f760g");
        f760gσ = coeffs.get("f760gs");
        f3 = coeffs.get("f3");
        f4 = coeffs.get("f4");
        f4mod = coeffs.get("f4mod");
        f5 = coeffs.get("f5");
        vc = coeffs.get("Vc");
        σc = coeffs.get("sig_c");
      }
    }

    SiteAmp_2018(Imt imt) {
      this(imt, 1.0);
    }

    SiteAmp_2018(Imt imt, double zScore) {
      c = new Coefficients(imt, COEFFS);
      this.zScore = zScore;
    }

    SiteTerm calc(double pgaRock, double vs30) {

      /*
       * Developer notes:
       *
       * Vs30 values outside the range 200 < Vs30 < 3000 m/s are clamped to the
       * supported range.
       *
       * ---------------------------------------------------------------------
       *
       * Comments from R implementation:
       *
       * This function is used to compute site amplification in ln units for
       * inputs Vs30 (in m/s) and PGAr (peak acceleration for 3000 m/s reference
       * rock in g). Outputs:
       *
       * 1) Linear and nonlinear site amplification (FLin, FNonlin)
       *
       * 2) Standard deviation (σLin, σNonlin)
       *
       * 3) Total site amplification (fT) and total standard deviation (σT)
       *
       * Notes:
       *
       * 1) We assume fv and f760 are independent, so variance of linear site
       * amplication is the sum of variance f760 and f760;
       *
       * 2) The third value of σc coefficient, which is used to compute
       * nonlinear site amplication. 0.12 in report, and 0.1 in excel. Here we
       * use 0.12.
       *
       * 3) Standard deviation of fv calculation is updated by adding a new
       * coefficient vf instead of using v1.
       */

      /* Vs30 filtering */

      if (vs30 >= V_MAX) {
        return new SiteTerm(0.0, 0.0, 0.0);
      } else if (vs30 < V_MIN) {
        vs30 = V_MIN;
      }

      /* Linear response */

      /*
       * Vs30 dependent f760 model: impedance vs. gradient
       *
       * The final model gives 2/3 weight to a weight scaling model of
       * [0.9i,0.1g] to [0.1i,0.9g] and 1/3 weight to a weight scaling model of
       * [0.5i,0.5g] to [0.1i,0.9g]. This results in a final weight scaling
       * model of [0.767i,0.233g] at 600 m/s and [0.1i,0.9g] at 400 m/s.
       */
      double wti = WT1; // impedance = 0.767 @ Vs30 > 600
      if (vs30 < VW2) {
        wti = WT2; // impedance = 0.1 @ Vs30 < 400
      } else if (vs30 < VW1) {
        wti = WT_SCALE * log(vs30 / VW2) + WT2;
      }
      double wtg = 1.0 - wti;

      double f760 = c.f760i * wti + c.f760g * wtg;
      double f760σ = c.f760iσ * wti + c.f760gσ * wtg;

      double fv = 0.0;
      if (vs30 <= c.v1) {
        fv = c.c * log(c.v1 / V_LIN_REF);
      } else if (vs30 <= c.v2) {
        fv = c.c * log(vs30 / V_LIN_REF);
      } else if (vs30 <= VU) {
        fv = c.c * log(c.v2 / V_LIN_REF);
      } else {
        /* Equivalent to equation 3 for 2000 < vs30 < 3000 */
        double f2000 = c.c * log(c.v2 / V_LIN_REF);
        fv = Interpolator.findY(log(VU), f2000, log(V_MAX), -f760, log(vs30));
      }

      double fvσ = 0.0;
      if (vs30 < c.vf) {
        double σT = c.σl - c.σvc;
        double vT = (vs30 - VL) / (c.vf - VL);
        fvσ = c.σl - 2.0 * σT * vT + σT * vT * vT;
      } else if (vs30 <= c.v2) {
        fvσ = c.σvc;
      } else if (vs30 <= VU) {
        double vT = (vs30 - c.v2) / (VU - c.v2);
        fvσ = c.σvc + (c.σu - c.σvc) * vT * vT;
      } else {
        fvσ = c.σu * (1.0 - log(vs30 / VU) / log(V_MAX / VU));
      }

      double fLin = fv + f760;
      double σLin = sqrt(fvσ * fvσ + f760σ * f760σ);

      /* Nonlinear response */

      double vRefNl = (c.imt.ordinal() >= Imt.SA0P4.ordinal()) ? V_MAX : V_LIN_REF;
      double rkRefTerm = log((pgaRock + c.f3) / c.f3);

      double fNonlin = 0.0;
      if (vs30 < c.vc) {
        double f2 = calcF4() * (exp(c.f5 * (min(vs30, vRefNl) - 360.0)) -
            exp(c.f5 * (vRefNl - 360.0)));
        fNonlin = f2 * rkRefTerm;
      }

      double σf2 = 0.0;
      if (vs30 < 300.0) {
        σf2 = c.σc;
      } else if (vs30 < 1000.0) {
        σf2 = c.σc - c.σc / log(1000.0 / 300.0) * log(vs30 / 300.0);
      }
      double σNonlin = σf2 * rkRefTerm;

      double fT = fLin + fNonlin;
      double σT = sqrt(σLin * σLin + σNonlin * σNonlin);

      return new SiteTerm(fT, σT, zScore);
    }

    double calcF4() {
      return c.f4;
    }
  }

  /*
   * Stewart et al. site amplification model, updated with alternate non-linear
   * coefficient f4 per final Hashash et al. (2020) model.
   */
  static class SiteAmp_2023 extends SiteAmp_2018 {

    SiteAmp_2023(Imt imt) {
      super(imt, Z_SCORE_5_95);
    }

    @Override
    double calcF4() {
      return c.f4 * 0.5 + c.f4mod * 0.5;
    }
  }

  /* Site amplification term and associated epistemic uncertainty. */
  static class SiteTerm {

    final double fs;
    final double σ;
    final double zScore;

    SiteTerm(double fs, double σ, double zScore) {
      this.fs = fs;
      this.σ = σ;
      this.zScore = zScore;
    }

    /*
     * Add site term to median ground motion, considering epistemic uncertainty.
     * Note that this method collapses three epistemic branches in ground motion
     * space. If these were carried through to hazard the branches would be
     * combined in rate space.
     */
    double apply(double μ) {
      if (fs == 0.0) {
        return μ;
      }
      double μAmp = μ + fs;
      return log(EPI_WTS[0] * exp(μAmp + σ * zScore) +
          EPI_WTS[1] * exp(μAmp) +
          EPI_WTS[2] * exp(μAmp - σ * zScore));
    }
  }

  static final double Z_CUT = 0.2;

  /*
   * Return the depth scaling factor for application of reference site scaling.
   * Method returns 0 for zSed = 0 or NaN. Value increases to 1.0 as zSed --> 1
   * km.
   *
   * @param z sediment depth (in km)
   */
  static double zSiteScale(double z) {
    if (Double.isNaN(z)) {
      return 0.0;
    }
    double s = 1.0 - Math.exp(-z / Z_CUT);
    return s * s * s * s;
  }
}
