package gov.usgs.earthquake.nshmp.model;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkCrustalDepth;
import static gov.usgs.earthquake.nshmp.Earthquakes.checkSlabDepth;
import static gov.usgs.earthquake.nshmp.Faults.checkStrike;
import static gov.usgs.earthquake.nshmp.fault.FocalMech.NORMAL;
import static gov.usgs.earthquake.nshmp.fault.FocalMech.REVERSE;
import static gov.usgs.earthquake.nshmp.fault.FocalMech.STRIKE_SLIP;
import static gov.usgs.earthquake.nshmp.model.PointSourceType.FIXED_STRIKE;
import static gov.usgs.earthquake.nshmp.model.SourceType.GRID;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NavigableMap;
import java.util.Optional;
import java.util.OptionalDouble;
import java.util.function.Function;
import java.util.function.Predicate;

import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Earthquakes;
import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.IntervalTable;
import gov.usgs.earthquake.nshmp.data.MutableXySequence;
import gov.usgs.earthquake.nshmp.data.Sequences;
import gov.usgs.earthquake.nshmp.data.XySequence;
import gov.usgs.earthquake.nshmp.fault.FocalMech;
import gov.usgs.earthquake.nshmp.fault.surface.RuptureScaling;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.LocationList;
import gov.usgs.earthquake.nshmp.geo.Locations;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.mfd.Mfd;
import gov.usgs.earthquake.nshmp.mfd.Mfds;
import gov.usgs.earthquake.nshmp.model.GridLoader.Key;
import gov.usgs.earthquake.nshmp.model.PointSource.DepthModel;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.tree.Trees;

/*
 * Gridded seismicity rupture set implementation.
 *
 * @author U.S. Geological Survey
 */
public class GridRuptureSet extends AbstractRuptureSet<PointSource> {

  private final SourceFeature feature;

  private final List<Location> locations;
  private final List<Mfd> mfds;
  private final LogicTree<List<Mfd>> mfdsTree;
  private final LogicTree<Mfd> mfdTree;

  private final List<Map<FocalMech, Double>> mechMaps; // may be nCopies
  private final boolean singularMechs;
  private final OptionalDouble strike;

  final RuptureScaling rupScaling;
  private final double maxDepth;
  private final PointSourceType pointSourceType;

  final DepthModel depthModel;
  // final boolean optimizable;

  final double gridSpacing;
  final int smoothingDensity;
  final double smoothingLimit;
  final OptionalDouble distanceBin;

  /*
   * Enforce expectation that all MFDs are GR*
   */

  GridRuptureSet(Builder builder) {
    super(builder);
    this.feature = builder.feature;

    this.locations = builder.locations;
    this.mfds = builder.mfds;
    this.mfdsTree = builder.mfdsTree;
    this.mfdTree = builder.mfdTree;

    this.rupScaling = builder.rupScaling;
    this.mechMaps = builder.mechMaps;
    this.singularMechs = builder.singularMechs;
    this.pointSourceType = builder.pointSourceType;
    this.strike = builder.strike;
    this.maxDepth = builder.maxDepth;

    double[] depthMags = this.mfds.get(0).data().xValues().toArray();

    this.depthModel = DepthModel.create(
        builder.magDepthMap,
        Doubles.asList(depthMags),
        OptionalDouble.of(maxDepth),
        OptionalDouble.empty());

    this.gridSpacing = builder.gridSpacing;
    this.smoothingDensity = builder.smoothingDensity;
    this.smoothingLimit = builder.smoothingLimit;
    this.distanceBin = builder.distanceBin;
  }

  @Override
  public int size() {
    return locations.size();
  }

  @Override
  public PointSource get(int index) {
    return getSource(index);
  }

  @Override
  public List<Feature> features() {
    return List.of(feature.source);
  }

  @Override
  public LogicTree<Mfd> mfdTree() {
    return mfdTree;
  }

  @Override
  public Location location(Location site) {
    throw new UnsupportedOperationException();
  }

  /**
   * The {@link PointSource} representation used by this source set.
   */
  public PointSourceType pointSourceType() {
    return pointSourceType;
  }

  /**
   * Whether this source set is capable of being optimized. There are
   * circumstances under which a grid optimization table can not be built.
   */
  public boolean optimizable() {
    return distanceBin.isPresent();
  }

  /**
   * For internal use only. Public for access outside of package.
   */
  public static String sizeString(RuptureSet<? extends Source> ruptures, int size) {
    if (ruptures instanceof Table) {
      Table t = (Table) ruptures;
      return t.parentCount() + " (" + t.rowCount + " of " + t.maximumSize + ")";
    }
    return Integer.toString(size);
  }

  @Override
  public Predicate<PointSource> distanceFilter(Location loc, double distance) {
    return new DistanceFilter(loc, distance);
  }

  /* Not inlined for use by area sources */
  static final class DistanceFilter implements Predicate<PointSource> {
    private final Predicate<Location> filter;

    DistanceFilter(Location loc, double distance) {
      filter = Locations.distanceAndRectangleFilter(loc, distance);
    }

    @Override
    public boolean test(PointSource source) {
      return filter.test(source.loc);
    }
  }

  @Override
  public Iterator<PointSource> iterator() {
    return new Iterator<PointSource>() {
      int caret = 0;

      @Override
      public boolean hasNext() {
        return caret < locations.size();
      }

      @Override
      public PointSource next() {
        return getSource(caret++);
      }

      @Override
      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
  }

  private PointSource getSource(int index) {

    /*
     * Stricter rules regarding what sorts of default mfds can be used with grid
     * sources (in an individual grid source set) will allow parsers to create
     * XySequence based MFDs directly using copyOf so as to not create zillions
     * of mag arrays.
     */
    Location loc = locations.get(index);
    Mfd mfd = mfds.get(index);
    Map<FocalMech, Double> mechMap = mechMaps.get(index);

    switch (pointSourceType) {
      case POINT:
        return new PointSource(this,
            GRID, loc, mfd, mechMap, rupScaling, depthModel, Optional.empty());

      case FINITE:
        return new PointSourceFinite(this,
            GRID, loc, mfd, mechMap, rupScaling, depthModel, Optional.empty());

      case FIXED_STRIKE:
        return new PointSourceFixedStrike(this,
            GRID, loc, mfd, mechMap, rupScaling, depthModel, strike.orElseThrow());

      default:
        throw new IllegalStateException("Unhandled point source type");
    }
  }

  static Builder builder() {
    return new Builder();
  }

  /* Single use builder. */
  static class Builder extends AbstractRuptureSet.Builder {

    private SourceFeature feature; // Grid or Zone
    private ModelData data;

    private List<Location> locations;
    private List<Mfd> mfds;
    private LogicTree<List<Mfd>> mfdsTree;
    private LogicTree<Mfd> mfdTree;

    // from GSSb
    private RuptureScaling rupScaling;
    private List<Map<FocalMech, Double>> mechMaps = Lists.newArrayList();
    private NavigableMap<Double, Map<Double, Double>> magDepthMap;
    private PointSourceType pointSourceType;
    private OptionalDouble strike = OptionalDouble.empty();
    private Double maxDepth;
    // private Double mMin;
    // private Double mMax;
    // private Double Δm;

    private Map<FocalMech, Double> mechMap;
    private boolean singularMechs = true;

    private Double gridSpacing;
    private Integer smoothingDensity;
    private Double smoothingLimit;
    private OptionalDouble distanceBin;

    private Builder() {
      super(SourceType.GRID);
    }

    Builder(SourceType type) {
      super(type);
    }

    @Override
    Builder name(String name) {
      super.name(name);
      return this;
    }

    @Override
    Builder id(int id) {
      super.id(id);
      return this;
    }

    private Builder gridConfig(SourceConfig.Grid gridConfig) {

      this.maxDepth = gridConfig.maxDepth;

      // makes bad assumption of GRID
      // reenable but need slab handling
      // validateDepth(this.maxDepth, SourceType.GRID);

      this.rupScaling = gridConfig.ruptureScaling;
      if (strike.isEmpty()) { // already set to FIXED_STRIKE
        this.pointSourceType = gridConfig.pointSourceType;
      }

      this.mechMap = convertFocalMechTree(gridConfig.focalMechTree);
      checkArgument(this.mechMap.size() == 3);

      this.magDepthMap = convertMagDepthMap(gridConfig.gridDepthMap);

      validateMagCutoffs(this.magDepthMap);

      // makes bad assumption of GRID
      // reenable but need slab handling
      // validateDepthMap(this.magDepthMap, SourceType.GRID);

      this.gridSpacing = gridConfig.gridSpacing;
      this.smoothingDensity = gridConfig.smoothingDensity;
      this.smoothingLimit = gridConfig.smoothingLimit;
      this.distanceBin = gridConfig.distanceBin;

      return this;
    }

    private Map<FocalMech, Double> convertFocalMechTree(LogicTree<FocalMech> tree) {
      Map<FocalMech, Double> fmMmap = new EnumMap<FocalMech, Double>(FocalMech.class);
      Arrays.stream(FocalMech.values()).forEach(v -> fmMmap.put(v, 0.0));
      tree.forEach(branch -> fmMmap.put(
          FocalMech.valueOf(branch.id()),
          branch.weight()));
      return Maps.immutableEnumMap(fmMmap);
    }

    private NavigableMap<Double, Map<Double, Double>> convertMagDepthMap(
        Map<String, GridDepthModel> depthMap) {
      return depthMap.values().stream()
          .collect(ImmutableSortedMap.toImmutableSortedMap(
              Comparator.naturalOrder(),
              model -> model.mRange.upperEndpoint(),
              model -> convertDepthTree(model.depthTree)));
    }

    private Map<Double, Double> convertDepthTree(LogicTree<Double> tree) {
      return tree.stream()
          .collect(ImmutableSortedMap.toImmutableSortedMap(
              Comparator.naturalOrder(),
              Branch::value,
              Branch::weight));
    }

    Builder feature(SourceFeature feature) {
      this.feature = feature;
      Properties props = feature.source.properties();
      OptionalDouble strike = props.getDouble(Key.STRIKE);
      if (strike.isPresent()) {
        checkStrike(strike.orElseThrow());
        this.strike = strike;
        this.pointSourceType = FIXED_STRIKE;
      }
      return this;
    }

    Builder modelData(ModelData data) {
      this.data = data;
      super.setting(data.tectonicSetting());
      super.gmms(data.gmms());
      gridConfig(data.sourceConfig().asGrid());
      return this;
    }

    Builder nodeData(
        List<Location> locations,
        List<Mfd> mfds,
        LogicTree<List<Mfd>> mfdsTree,
        Optional<List<Map<FocalMech, Double>>> mechMaps) {

      checkNotNull(locations);
      checkNotNull(mfds);
      checkNotNull(mfdsTree);
      checkNotNull(mechMaps);
      checkArgument(locations.size() > 0);
      checkArgument(mfds.size() == locations.size());
      mfdsTree.forEach(branch -> checkArgument(branch.value().size() == locations.size()));

      if (mechMaps.isPresent()) {
        checkArgument(mechMaps.orElseThrow().size() == locations.size());
        this.mechMaps = mechMaps.orElseThrow();
      }

      this.locations = locations;
      this.mfds = mfds;
      this.mfdsTree = mfdsTree;
      this.mfdTree = Trees.transform(mfdsTree, Mfds::combine, Optional.empty());

      return this;
    }

    static void validateDepthMap(
        Map<Double, Map<Double, Double>> magDepthMap,
        SourceType type) {
      for (Map<Double, Double> magMap : magDepthMap.values()) {
        for (double depth : magMap.keySet()) {
          validateDepth(depth, type);
        }
      }
    }

    static void validateDepth(double depth, SourceType type) {
      switch (type) {
        case GRID:
          checkCrustalDepth(depth);
          break;
        case SLAB:
          checkSlabDepth(depth);
          break;
        case ZONE:
          checkCrustalDepth(depth);
          break;
        default:
          throw new IllegalStateException(type + " not a grid or related source type");
      }
    }

    static void validateMaxAndMapDepths(Map<Double, Map<Double, Double>> magDepthMap,
        double maxDepth, String id) {
      for (Map<Double, Double> magMap : magDepthMap.values()) {
        for (double depth : magMap.keySet()) {
          checkState(depth <= maxDepth, "%s mag-depth-weight map depth %s > %s", id,
              depth, maxDepth);
        }
      }
    }

    static void validateMagCutoffs(Map<Double, Map<Double, Double>> magDepthMap) {
      double mMax = Earthquakes.MAG_RANGE.upperEndpoint();
      for (double mag : magDepthMap.keySet()) {
        if (mag > mMax) {
          return;
        }
      }
      throw new IllegalStateException("MagDepthMap must contain at least one M > " + mMax);
    }

    void validateAndInit(String label) {

      validateState(label);

      checkNotNull(feature, "%s feature", label);
      checkNotNull(data, "%s MFD reference data", label);

      checkNotNull(locations, "%s locations", label);
      checkNotNull(mfds, "%s total MFDs", label);
      checkState(!mfds.isEmpty(), "%s has no MFDs", label);
      checkNotNull(mfdsTree, "%s MFDs logic tree", label);
      checkNotNull(mfdTree, "%s collapsed MFD logic tree", label);

      checkNotNull(pointSourceType != null, "%s point source type", label);
      checkNotNull(rupScaling, "%s rupture-scaling relation", label);
      checkNotNull(magDepthMap, "%s mag-depth-weight map", label);
      checkNotNull(maxDepth, "%s max depth", label);
      checkNotNull(mechMap, "%s focal mech map", label);

      checkNotNull(gridSpacing, "%s grid spacing", label);
      checkNotNull(smoothingDensity != null, "%s smoothing density", label);
      checkNotNull(smoothingLimit != null, "%s smoothing limit", label);
      checkNotNull(distanceBin != null, "%s optimization table distance bin", label);

      // checkState(mMin != null, "%s min mag not set", buildId);
      // checkState(mMax != null, "%s max mag not set", buildId);
      // checkState(Δm != null, "%s delta mag not set", buildId);

      /*
       * There are too many assumptions built into this; whose to say ones bin
       * spacing should be only be in the hundredths?
       *
       * Where did this come from anyway? Are mag deltas really all that strange
       *
       * We should read precision of supplied mMin and mMax and delta and use
       * largest for formatting
       *
       * In the case of single combined/flattened MFDs, mags may not be
       * uniformly spaced. Can this be refactored
       */

      /*
       * Validate size of mechMaps; size could get out of sync if mixed calls to
       * location(...) were made; one can imagine a future use case where a
       * default is required with an override in a few locations; for now, if
       * custom mechMaps are required, there must be one for each node. If no
       * custom maps supplied populate mechMaps with nCopies (singleton list
       * with multiple elements)
       */

      if (mechMaps.isEmpty()) {
        mechMaps = Collections.nCopies(locations.size(), mechMap);
      } else {
        checkState(
            mechMaps.size() == locations.size(),
            "%s only %s of %s focal mech maps were added", label,
            mechMaps.size(), locations.size());
        singularMechs = false;
      }

      /*
       * Validate depths. depths will already have been checked for consistency
       * with allowable depths for different source types. Must also ensure that
       * all depths (zTor) in the magDepthMap are < maxDepth.
       */
      validateMaxAndMapDepths(magDepthMap, maxDepth, label);

      /* Validate strike value and type agreement. */
      if (strike.isPresent()) {
        checkState(pointSourceType == FIXED_STRIKE,
            "Source type must be FIXED_STRIKE for strike [%s]", strike.orElseThrow());
      } else {
        checkState(pointSourceType != FIXED_STRIKE,
            "Source type FIXED_STRIKE invalid for strike [%s]", strike);
      }

    }

    /* Create a new rupture set. */
    GridRuptureSet build() {
      validateAndInit("GridRuptureSet.Builder");
      return new GridRuptureSet(this);
    }
  }

  /*
   * Notes on refactoring GridRuptureSet xml.
   *
   * The existing implementation permits one mfd per node and that maps to the
   * default of the same 'type'. This was fine assuming mfds were collapsed
   * (CEUS 2008) or MFDs were explicitly one type or another as required for CA
   * 2008 grids (GR or INCR, the latter to handle the M>6.5 rate reduction).
   *
   * However, we want to be able to track mMax and other logic tree branches.
   *
   * Given gridded sources where rate or a-value at each node is consistent for
   * all applicable MFDs:
   *
   * - each node will map to all defaults
   *
   * - defaults must have unique 'id' (logic tree branch id), but can be of
   * similar 'type'; throw error if field doesn't match
   *
   * - CA 2008 will need to be converted to only INCR MFDs with explicit rates
   * for all nodes.
   *
   * - refactor IncrementalMfd to Mfd in XML
   *
   * - refactor Node to Source
   *
   * - usually the MFD 'type' is consistent across branches for a given grid
   * source set so collapsing MFD's isn't too problematic.
   *
   * Grid optimizations:
   *
   * - create GridRuptureSet subclass, one for each default MFD (siblings)
   *
   * - also create a subclass that uses a collapsed MFD, or perhaps this is just
   * the original.
   */

  /*
   * If, for a basic HazardResult, we want to be able to give a per-source-set
   * decomposition by ground motion model, or just a decomposition of the total
   * curve, we'll need to have a table of the curves for every model.
   *
   * If not necessary, then can have table of total curves and table of mean
   * (and sigma?) for each model. Just mean is necessary for disaggeregation
   * epsilon
   *
   * OK... so...
   *
   * Preliminary implementations of grid source optimizations modeled after the
   * NSHMP Fortran codes porecomputed median curves in distance and magnitude
   * (using a weighted combination of Gmms) and then performed lookups for each
   * source, aggregating a total curve along the way. This approach is lossy in
   * that data for individual Gmms is lost, and it was never extended to support
   * disaggregation where ground motion mean and sigma are required.
   *
   * Further consideration of these issues suggests that, rather than
   * aggregating curves along the way, we should build a separate table in
   * magnitude and distance of rates while looping over sources. In the end,
   * curves could be computed once for each distance and magnitude bin. Although
   * full curves for each Gmm could be precomputed, the time to loop over the
   * rate table may not be significant enough to warrant the memory overhead
   * (bear in mind that's a lot of curves when considering large logic trees of
   * Gmms and numerous periods).
   *
   * There is also the additional issue of additional epistemic uncertinaty on
   * ground motions, which does not need to be considered here if building
   * magnitude-distance rate tables.
   *
   * There is the additional issue of different focal mechanisms. For NGAW2 and
   * the WUS, we would need to have 5 curves per gmm and r/m bin: 2 reverse, 2
   * normal 1 strike slip
   *
   * Also assumes uniform vs30 or bsain terms; these will likely be spatially
   * varying in the future; this would also be incompatible with gmm tables.
   *
   * Precomputed curves may still be warranted for map calculations where Gmm
   * specific data and disaggregation are irrelevant.
   */

  /*
   * Why, you say?
   *
   * Simply put, speed. In the 2014 CEUS NSHM, 1000km from a site nets about 30k
   * sources, each of which has an associated MFD with up to 33 values (and that
   * assumes the different mMax models have been collapsed together). So 990k
   * curve calculations per GMM. However, if the rates of those sources are
   * first aggregated into a matrix of distance (300) and magnitude (33) bins,
   * then only 900 chazard curve calculations need be performed per GMM. Ha!
   */

  /*
   * need to specify magnitude and distance discretization
   */

  /**
   * Create a {@code Function} for a location the condenses a
   * {@code GridRuptureSet} into tabular form (distance, magnitude and azimuth
   * bins) for speedier iteration.
   *
   * @param loc reference point for table
   */
  public static Function<GridRuptureSet, RuptureSet<? extends Source>> optimizer(
      Location loc, boolean smooth) {
    return new Optimizer(loc, smooth);
  }

  private static class Optimizer implements
      Function<GridRuptureSet, RuptureSet<? extends Source>> {

    private final Location loc;
    private final boolean smooth;

    Optimizer(Location loc, boolean smooth) {
      this.loc = loc;
      this.smooth = smooth;
    }

    @Override
    public Table apply(GridRuptureSet sources) {
      return new Table(sources, loc, smooth);
    }
  }

  /*
   * Notes on dealing with mixedMech situations (e.g. UC3)
   *
   * Need 3 tables (SS, R, N)
   *
   * Rate in each R-M bin gets partitioned across three tables.
   *
   * When building sources we could normalize the rates to create a mechMap on
   * the fly and sum back to the total rate.
   *
   * Alternatively, and preferably, we reconsolidate partitioned rates into a
   * list
   */

  /*
   * Upgrade this to DataVolume to handle azimuth bins?? split over focal mechs?
   * required for UC3 grids
   */

  /**
   * Tabular implementation of a {@code GridRuptureSet}. This class consolidates
   * the point sources that influence hazard at a site using a
   * magnitude-distance-rate {@code DataTable}, from which a list of sources is
   * generated. A {@code Table} is created on a per-calculation basis and is
   * unique to a location.
   *
   * @see GridRuptureSet#optimizer(Location)
   */
  private static final class Table extends AbstractRuptureSet<PointSource> {

    private final GridRuptureSet parent;
    private final Location origin;
    private final List<PointSource> sources;

    /*
     * Row count reflects the number of rows used in a DataTable when building
     * sources. Although this will likely be the same as sources.size(), it may
     * not be. For example, when using multi-mechs many more sources are created
     * because the different focal mechs are (can) not be combined given
     * possibly varying rates across different magnitudes.
     */
    private int rowCount;
    private int maximumSize;
    private int parentCount;

    private Table(GridRuptureSet parent, Location origin, boolean smoothed) {

      // This is dirty - arose from conversion to ass constructor taking
      // builder
      super(new AbstractRuptureSet.Builder(parent.type()) {
        {
          this.name = parent.name();
          this.id = parent.id();
          this.setting = parent.setting();
          this.gmmTree = parent.gmmTree();
        }
      });

      this.parent = parent;
      setLeafWeight(parent.weight());
      this.origin = origin;
      this.sources = parent.singularMechs
          ? initSources(smoothed)
          : initMultiMechSources(smoothed);
    }

    /**
     * The number of sources drawn from a parent {@code GridRuptureSet} during
     * initialization.
     */
    public int parentCount() {
      return parentCount;
    }

    @Override
    public String name() {
      return parent.name() + " (opt)";
    }

    @Override
    public int size() {
      return parent.size();
    }

    @Override
    public PointSource get(int index) {
      return sources.get(index);
    }

    @Override
    public List<Feature> features() {
      return List.of(parent.feature.source);
    }

    @Override
    public Predicate<PointSource> distanceFilter(Location loc, double distance) {
      throw new UnsupportedOperationException();
    }

    @Override
    public Iterator<PointSource> iterator() {
      throw new UnsupportedOperationException();
    }

    @Override
    public LogicTree<Mfd> mfdTree() {
      throw new UnsupportedOperationException();
    }

    @Override
    public Location location(Location site) {
      throw new UnsupportedOperationException();
    }

    @Override
    public Iterable<PointSource> iterableForLocation(Location loc) {
      /*
       * Ignore location; simply iterate over the list of sources. Source list
       * will be empty if mfdTable is empty (all zeros).
       */
      return sources;
    }

    private static final double SRC_TO_SITE_AZIMUTH = 0.0;
    // private static final double SMOOTHING_LIMIT = 40.0; // km

    /* creates the type of point source specified in the parent */
    private List<PointSource> initSources(boolean smoothed) {
      /* For now, should only be getting here for GR MFDs */
      Mfd modelMfd = parent.mfds.get(0);
      double[] mags = modelMfd.data().xValues().toArray();

      // Can we get this from the mfd.properties?
      // the original min max won't have offsets
      double Δm = mags[1] - mags[0];
      double ΔmBy2 = Δm / 2.0;
      double mMin = mags[0] - ΔmBy2;
      double mMax = mags[mags.length - 1] + ΔmBy2;

      double rMax = parent.gmmTree().maxDistance().orElseThrow();
      double distanceBin = parent.distanceBin.orElseThrow();
      double[] smoothingOffsets = smoothingOffsets(
          parent.smoothingDensity,
          parent.gridSpacing);

      IntervalTable.Builder tableBuilder = new IntervalTable.Builder()
          .rows(0.0, rMax, distanceBin)
          .columns(mMin, mMax, Δm);

      for (PointSource source : parent.iterableForLocation(origin)) {
        double r = Locations.horzDistanceFast(origin, source.loc);
        /* Experimental smoothing. */
        if (smoothed && r < parent.smoothingLimit) {
          addSmoothed(tableBuilder, origin, source.loc, source.mfd, smoothingOffsets);
        } else {
          tableBuilder.add(r, source.mfd.data());
        }
        parentCount++;
      }
      IntervalTable mfdTable = tableBuilder.build();

      // System.out.println(parent.name());
      // System.out.println(mfdTable);
      // System.out.println("TableSum: " + mfdTable.collapse().sum());
      // System.out.println(mfdTable.rows());

      List<Double> distances = mfdTable.rows();
      maximumSize = distances.size();
      List<PointSource> b = new ArrayList<>();
      for (double r : distances) {
        Mfd mfd = Mfd.create(mfdTable.row(r));
        if (mfd.data().isClear()) {
          continue;
        }
        Location loc = Locations.location(origin, SRC_TO_SITE_AZIMUTH, r);
        b.add(new PointSourceFinite(
            parent,
            parent.type(),
            loc,
            mfd,
            parent.mechMaps.get(0),
            parent.rupScaling,
            parent.depthModel,
            Optional.empty()));
        rowCount++;
      }
      return List.copyOf(b);
    }

    /* always creates finite point sources */
    private List<PointSource> initMultiMechSources(boolean smoothed) {

      /* For now, should only be getting here for GR MFDs */
      Mfd modelMfd = parent.mfds.get(0);
      double[] mags = modelMfd.data().xValues().toArray();

      double Δm = mags[1] - mags[0];
      double ΔmBy2 = Δm / 2.0;
      double mMin = mags[0] - ΔmBy2;
      double mMax = mags[mags.length - 1] + ΔmBy2;

      double rMax = parent.gmmTree().maxDistance().orElseThrow();
      double distanceBin = parent.distanceBin.orElseThrow();
      double[] smoothingOffsets = smoothingOffsets(
          parent.smoothingDensity,
          parent.gridSpacing);

      IntervalTable.Builder ssTableBuilder = new IntervalTable.Builder()
          .rows(0.0, rMax, distanceBin)
          .columns(mMin, mMax, Δm);

      IntervalTable.Builder rTableBuilder = new IntervalTable.Builder()
          .rows(0.0, rMax, distanceBin)
          .columns(mMin, mMax, Δm);

      IntervalTable.Builder nTableBuilder = new IntervalTable.Builder()
          .rows(0.0, rMax, distanceBin)
          .columns(mMin, mMax, Δm);

      for (PointSource source : parent.iterableForLocation(origin)) {
        double r = Locations.horzDistanceFast(origin, source.loc);

        Mfd ssMfd = Mfd.Builder.from(source.mfd).scale(source.mechWtMap.get(STRIKE_SLIP)).build();
        Mfd rMfd = Mfd.Builder.from(source.mfd).scale(source.mechWtMap.get(REVERSE)).build();
        Mfd nMfd = Mfd.Builder.from(source.mfd).scale(source.mechWtMap.get(NORMAL)).build();

        if (smoothed && r < parent.smoothingLimit) {
          addSmoothed(ssTableBuilder, origin, source.loc, ssMfd, smoothingOffsets);
          addSmoothed(rTableBuilder, origin, source.loc, rMfd, smoothingOffsets);
          addSmoothed(nTableBuilder, origin, source.loc, nMfd, smoothingOffsets);
        } else {
          ssTableBuilder.add(r, ssMfd.data());
          rTableBuilder.add(r, rMfd.data());
          nTableBuilder.add(r, nMfd.data());
        }
        parentCount++;
      }

      IntervalTable ssTable = ssTableBuilder.build();
      // System.out.println("SS Table:" + Text.NEWLINE + ssTable);
      IntervalTable rTable = rTableBuilder.build();
      // System.out.println("R Table:" + Text.NEWLINE + rTable);
      IntervalTable nTable = nTableBuilder.build();
      // System.out.println("N Table:" + Text.NEWLINE + nTable);

      // DataTable tableSum = DataTable.Builder.fromModel(ssTable)
      // .add(ssTable)
      // .add(rTable)
      // .add(nTable)
      // .build();

      // XySequence tableMfdSum =
      // XySequence.emptyCopyOf(tableSum.row(0.1));
      // for (double row : tableSum.rows()) {
      // tableMfdSum.add(tableSum.row(row));
      // }

      // System.out.println("tableMfd:");
      // System.out.println(tableMfdSum);

      List<Double> distances = ssTable.rows();
      maximumSize = distances.size();
      List<PointSource> b = new ArrayList<>();
      for (double r : distances) {
        Location loc = Locations.location(origin, SRC_TO_SITE_AZIMUTH, r);
        boolean tableRowUsed = false;

        Mfd ssMfd = Mfd.create(ssTable.row(r));
        if (ssMfd.data().y(0) > 0.0 || !ssMfd.data().isClear()) {
          b.add(new PointSourceFinite(
              parent,
              parent.type(),
              loc,
              ssMfd,
              parent.mechMaps.get(0),
              parent.rupScaling,
              parent.depthModel,
              Optional.empty()));
          tableRowUsed = true;
        }

        Mfd rMfd = Mfd.create(rTable.row(r));
        if (rMfd.data().y(0) > 0.0 || !rMfd.data().isClear()) {
          b.add(new PointSourceFinite(
              parent,
              parent.type(),
              loc,
              rMfd,
              parent.mechMaps.get(0),
              parent.rupScaling,
              parent.depthModel,
              Optional.empty()));
          tableRowUsed = true;
        }

        Mfd nMfd = Mfd.create(nTable.row(r));
        if (nMfd.data().y(0) > 0.0 || !nMfd.data().isClear()) {
          b.add(new PointSourceFinite(
              parent,
              parent.type(),
              loc,
              nMfd,
              parent.mechMaps.get(0),
              parent.rupScaling,
              parent.depthModel,
              Optional.empty()));
          tableRowUsed = true;
        }

        if (tableRowUsed) {
          rowCount++;
        }
      }
      return List.copyOf(b);
    }

    private static double[] smoothingOffsets(int density, double spacing) {
      return offsets(spacing, (density == 4)
          ? OFFSET_SCALE_4
          : OFFSET_SCALE_10);
    }
  }

  static void addSmoothed(
      IntervalTable.Builder builder,
      Location site,
      Location source,
      Mfd mfd,
      double[] smoothingOffsets) {

    LocationList distributedLocs = distributedLocations(source, smoothingOffsets);
    double rateScale = 1.0 / distributedLocs.size();
    XySequence scaledMfd = MutableXySequence.copyOf(mfd.data()).multiply(rateScale);
    for (Location loc : distributedLocs) {
      double r = Locations.horzDistanceFast(site, loc);
      builder.add(r, scaledMfd);
    }
  }

  static LocationList distributedLocations(Location loc, double[] offsets) {
    double lon = loc.longitude;
    double lat = loc.latitude;
    LocationList.Builder locations = LocationList.builder();
    for (double Δlat : offsets) {
      double offsetLat = lat + Δlat;
      for (double Δlon : offsets) {
        double offsetLon = lon + Δlon;
        locations.add(Location.create(offsetLon, offsetLat));
      }
    }
    return locations.build();
  }

  private static final int PRECISION = 5;
  private static final double[] OFFSET_SCALE_10 = Sequences
      .arrayBuilder(-0.45, 0.45, 0.1)
      .centered()
      .build();
  private static final double[] OFFSET_SCALE_4 = Sequences
      .arrayBuilder(-0.375, 0.375, 0.25)
      .centered()
      .build();

  /*
   * Create lat-lon offset values for use when building list of distributed
   * point source locations.
   *
   * NOTE DataArray would be cleaner here
   */
  private static double[] offsets(double cellScale, double[] offsetArray) {
    return DoubleData.round(PRECISION, DoubleData.multiply(
        cellScale,
        Arrays.copyOf(offsetArray, offsetArray.length)));
  }
}
